"""
Integration tests for the vault modules
"""

import logging

import pytest
from saltfactories.utils import random_string

# pylint: disable=unused-import
from tests.support.pytest.vault import (
    vault_container_version,
    vault_delete_secret,
    vault_environ,
    vault_list_secrets,
    vault_write_secret,
)

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd", "vault", "getent"),
    pytest.mark.usefixtures("vault_container_version"),
]


@pytest.fixture(scope="class")
def pillar_tree(vault_salt_master, vault_salt_minion):
    top_file = f"""
    base:
      '{vault_salt_minion.id}':
        - sdb
    """
    sdb_pillar_file = """
    test_vault_pillar_sdb: sdb://sdbvault/secret/test/test_pillar_sdb/foo
    """
    top_tempfile = vault_salt_master.pillar_tree.base.temp_file("top.sls", top_file)
    sdb_tempfile = vault_salt_master.pillar_tree.base.temp_file(
        "sdb.sls", sdb_pillar_file
    )

    with top_tempfile, sdb_tempfile:
        yield


@pytest.fixture(scope="class")
def vault_master_config(vault_port):
    return {
        "open_mode": True,
        "peer_run": {
            ".*": [
                "vault.get_config",
                "vault.generate_new_token",
            ],
        },
        "vault": {
            "auth": {
                "token": "testsecret",
            },
            "issue": {
                "token": {
                    "params": {
                        "num_uses": 0,
                    }
                }
            },
            "policies": {
                "assign": [
                    "salt_minion",
                ]
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        },
        "minion_data_cache": True,
    }


@pytest.fixture(scope="class")
def vault_salt_master(salt_factories, vault_port, vault_master_config):
    factory = salt_factories.salt_master_daemon(
        "vault-sdbmaster", defaults=vault_master_config
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="class")
def sdb_profile():
    return {}


@pytest.fixture(scope="class")
def vault_salt_minion(vault_salt_master, sdb_profile):
    assert vault_salt_master.is_running()
    config = {"open_mode": True, "grains": {}, "sdbvault": {"driver": "vault"}}
    config["sdbvault"].update(sdb_profile)
    factory = vault_salt_master.salt_minion_daemon(
        random_string("vault-sdbminion", uppercase=False),
        defaults=config,
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="class")
def vault_salt_call_cli(vault_salt_minion):
    return vault_salt_minion.salt_call_cli()


@pytest.fixture(scope="class")
def vault_salt_run_cli(vault_salt_master):
    return vault_salt_master.salt_run_cli()


@pytest.fixture
def kv_root_dual_item(vault_container_version):
    if vault_container_version == "latest":
        vault_write_secret("salt/user1", password="p4ssw0rd", desc="test user")
        vault_write_secret("salt/user/user1", password="p4ssw0rd", desc="test user")
    yield
    if vault_container_version == "latest":
        vault_delete_secret("salt/user1")
        vault_delete_secret("salt/user/user1")


@pytest.mark.parametrize("vault_container_version", ["1.3.1", "latest"], indirect=True)
def test_sdb_kv_kvv2_path_local(salt_call_cli, vault_container_version):
    ret = salt_call_cli.run(
        "--local",
        "sdb.set",
        uri="sdb://sdbvault/kv-v2/test/test_sdb_local/foo",
        value="local",
    )
    assert ret.returncode == 0
    assert ret.data is True
    ret = salt_call_cli.run(
        "--local", "sdb.get", "sdb://sdbvault/kv-v2/test/test_sdb_local/foo"
    )
    assert ret.data
    assert ret.data == "local"


@pytest.mark.usefixtures("kv_root_dual_item")
@pytest.mark.parametrize("vault_container_version", ["latest"], indirect=True)
def test_sdb_kv_dual_item(salt_call_cli, vault_container_version):
    ret = salt_call_cli.run("--local", "sdb.get", "sdb://sdbvault/salt/data/user1")
    assert ret.data
    assert ret.data == {"desc": "test user", "password": "p4ssw0rd"}


def test_sdb_runner(salt_run_cli):
    ret = salt_run_cli.run(
        "sdb.set", uri="sdb://sdbvault/secret/test/test_sdb_runner/foo", value="runner"
    )
    assert ret.returncode == 0
    assert ret.data is True
    ret = salt_run_cli.run(
        "sdb.get", uri="sdb://sdbvault/secret/test/test_sdb_runner/foo"
    )
    assert ret.returncode == 0
    assert ret.stdout
    assert ret.stdout == "runner"


@pytest.mark.usefixtures("pillar_tree")
class TestSDB:
    def test_sdb(self, vault_salt_call_cli):
        ret = vault_salt_call_cli.run(
            "sdb.set", uri="sdb://sdbvault/secret/test/test_sdb/foo", value="bar"
        )
        assert ret.returncode == 0
        assert ret.data is True
        ret = vault_salt_call_cli.run(
            "sdb.get", uri="sdb://sdbvault/secret/test/test_sdb/foo"
        )
        assert ret.returncode == 0
        assert ret.data
        assert ret.data == "bar"

    def test_config(self, vault_salt_call_cli):
        ret = vault_salt_call_cli.run(
            "sdb.set", uri="sdb://sdbvault/secret/test/test_pillar_sdb/foo", value="baz"
        )
        assert ret.returncode == 0
        assert ret.data is True
        ret = vault_salt_call_cli.run("config.get", "test_vault_pillar_sdb")
        assert ret.returncode == 0
        assert ret.data
        assert ret.data == "baz"


class TestGetOrSetHashSingleUseToken:
    @pytest.fixture(scope="class")
    def vault_master_config(self, vault_port):
        return {
            "open_mode": True,
            "peer_run": {
                ".*": [
                    "vault.get_config",
                    "vault.generate_new_token",
                ],
            },
            "vault": {
                "auth": {"token": "testsecret"},
                "cache": {
                    "backend": "file",
                },
                "issue": {
                    "type": "token",
                    "token": {
                        "params": {
                            "num_uses": 1,
                        }
                    },
                },
                "policies": {
                    "assign": [
                        "salt_minion",
                    ],
                },
                "server": {
                    "url": f"http://127.0.0.1:{vault_port}",
                },
            },
            "minion_data_cache": True,
        }

    @pytest.fixture
    def get_or_set_absent(self):
        secret_path = "secret/test"
        secret_name = "sdb_get_or_set_hash"
        ret = vault_list_secrets(secret_path)
        if secret_name in ret:
            vault_delete_secret(f"{secret_path}/{secret_name}")
        ret = vault_list_secrets(secret_path)
        assert secret_name not in ret
        try:
            yield
        finally:
            ret = vault_list_secrets(secret_path)
            if secret_name in ret:
                vault_delete_secret(f"{secret_path}/{secret_name}")

    @pytest.mark.usefixtures("get_or_set_absent")
    @pytest.mark.parametrize(
        "vault_container_version", ["1.3.1", "latest"], indirect=True
    )
    def test_sdb_get_or_set_hash_single_use_token(self, vault_salt_call_cli):
        """
        Test that sdb.get_or_set_hash works with uses=1.
        This fails for versions that do not have the sys/internal/ui/mounts/:path
        endpoint (<0.10.0) because the path metadata lookup consumes a token use there.
        Issue #60779
        """
        ret = vault_salt_call_cli.run(
            "sdb.get_or_set_hash",
            "sdb://sdbvault/secret/test/sdb_get_or_set_hash/foo",
            10,
        )
        assert ret.returncode == 0
        result = ret.data
        assert result
        ret = vault_salt_call_cli.run(
            "sdb.get_or_set_hash",
            "sdb://sdbvault/secret/test/sdb_get_or_set_hash/foo",
            10,
        )
        assert ret.returncode == 0
        assert ret.data
        assert ret.data == result


class TestSDBSetPatch:
    @pytest.fixture(scope="class")
    def sdb_profile(self):
        return {"patch": True}

    def test_sdb_set(self, vault_salt_call_cli):
        # Write to an empty path
        ret = vault_salt_call_cli.run(
            "sdb.set", uri="sdb://sdbvault/secret/test/test_sdb_patch/foo", value="bar"
        )
        assert ret.returncode == 0
        assert ret.data is True
        # Write to an existing path, this should not overwrite the previous key
        ret = vault_salt_call_cli.run(
            "sdb.set", uri="sdb://sdbvault/secret/test/test_sdb_patch/bar", value="baz"
        )
        assert ret.returncode == 0
        assert ret.data is True
        # Ensure the first value is still there
        ret = vault_salt_call_cli.run(
            "sdb.get", uri="sdb://sdbvault/secret/test/test_sdb_patch/foo"
        )
        assert ret.returncode == 0
        assert ret.data
        assert ret.data == "bar"
        # Ensure the second value was written
        ret = vault_salt_call_cli.run(
            "sdb.get", uri="sdb://sdbvault/secret/test/test_sdb_patch/bar"
        )
        assert ret.returncode == 0
        assert ret.data
        assert ret.data == "baz"



import sys

import salt.client.ssh
import salt.utils.parsers


class SaltSSH(salt.utils.parsers.SaltSSHOptionParser):
    """
    Used to Execute the salt ssh routine
    """

    def run(self):
        if "-H" in sys.argv or "--hosts" in sys.argv:
            sys.argv += ["x", "x"]  # Hack: pass a mandatory two options
            # that won't be used anyways with -H or --hosts
        self.parse_args()

        ssh = salt.client.ssh.SSH(self.config)
        try:
            ssh.run()
        finally:
            ssh.fsclient.destroy()



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
    :codeauthor: Joe Julian <me@joejulian.name>

    Test cases for salt.modules.glusterfs
"""

import pytest

import salt.modules.glusterfs as glusterfs
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock, patch


class GlusterResults:
    """This class holds the xml results from gluster cli transactions"""

    class v34:
        """This is for version 3.4 results"""

        class list_peers:
            """results from "peer status" """

        class peer_probe:
            fail_cant_connect = fail_bad_hostname = "\n".join(
                [
                    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>',
                    "<cliOutput>",
                    "  <opRet>-1</opRet>",
                    "  <opErrno>107</opErrno>",
                    "  <opErrstr>Probe returned with unknown errno 107</opErrstr>",
                    "</cliOutput>",
                    "",
                ]
            )

            success_self = "\n".join(
                [
                    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                    "  <cliOutput>",
                    "  <opRet>0</opRet>",
                    "  <opErrno>1</opErrno>",
                    "  <opErrstr>(null)</opErrstr>",
                    "  <output>success: on localhost not needed</output>",
                    "</cliOutput>",
                    "",
                ]
            )
            success_other = "\n".join(
                [
                    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                    "  <cliOutput>",
                    "  <opRet>0</opRet>",
                    "  <opErrno>0</opErrno>",
                    "  <opErrstr>(null)</opErrstr>",
                    "  <output>success</output>",
                    "</cliOutput>",
                    "",
                ]
            )
            success_hostname_after_ip = success_other
            success_ip_after_hostname = success_other
            success_already_peer = {
                "ip": "\n".join(
                    [
                        '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                        "  <cliOutput>",
                        "  <opRet>0</opRet>",
                        "  <opErrno>2</opErrno>",
                        "  <opErrstr>(null)</opErrstr>",
                        "  <output>success: host 10.0.0.2 port 24007 already in peer"
                        " list</output>",
                        "</cliOutput>",
                        "",
                    ]
                ),
                "hostname": "\n".join(
                    [
                        '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                        "  <cliOutput>",
                        "  <opRet>0</opRet>",
                        "  <opErrno>2</opErrno>",
                        "  <opErrstr>(null)</opErrstr>",
                        "  <output>success: host server2 port 24007 already in peer"
                        " list</output>",
                        "</cliOutput>",
                        "",
                    ]
                ),
            }
            success_reverse_already_peer = {
                "ip": "\n".join(
                    [
                        '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                        "  <cliOutput>",
                        "  <opRet>0</opRet>",
                        "  <opErrno>2</opErrno>",
                        "  <opErrstr>(null)</opErrstr>",
                        "  <output>success: host 10.0.0.1 port 24007 already in peer"
                        " list</output>",
                        "</cliOutput>",
                        "",
                    ]
                ),
                "hostname": "\n".join(
                    [
                        '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                        "  <cliOutput>",
                        "  <opRet>0</opRet>",
                        "  <opErrno>2</opErrno>",
                        "  <opErrstr>(null)</opErrstr>",
                        "  <output>success: host server1 port 24007 already in peer"
                        " list</output>",
                        "</cliOutput>",
                        "",
                    ]
                ),
            }
            success_first_hostname_from_second_first_time = success_other
            success_first_hostname_from_second_second_time = (
                success_reverse_already_peer["hostname"]
            )
            success_first_ip_from_second_first_time = success_reverse_already_peer["ip"]

    class v37:
        class peer_probe:
            fail_cant_connect = fail_bad_hostname = "\n".join(
                [
                    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>',
                    "<cliOutput>",
                    "  <opRet>-1</opRet>",
                    "  <opErrno>107</opErrno>",
                    "  <opErrstr>Probe returned with Transport endpoint is not"
                    " connected</opErrstr>",
                    "</cliOutput>",
                    "",
                ]
            )
            success_self = "\n".join(
                [
                    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                    "  <cliOutput>",
                    "  <opRet>0</opRet>",
                    "  <opErrno>1</opErrno>",
                    "  <opErrstr/>",
                    "  <output>Probe on localhost not needed</output>",
                    "</cliOutput>",
                    "",
                ]
            )
            success_other = "\n".join(
                [
                    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                    "  <cliOutput>",
                    "  <opRet>0</opRet>",
                    "  <opErrno>0</opErrno>",
                    "  <opErrstr/>",
                    "  <output/>",
                    "</cliOutput>",
                    "",
                ]
            )
            success_hostname_after_ip = success_other
            success_ip_after_hostname = success_other
            success_already_peer = {
                "ip": "\n".join(
                    [
                        '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                        "  <cliOutput>",
                        "  <opRet>0</opRet>",
                        "  <opErrno>2</opErrno>",
                        "  <opErrstr/>",
                        "  <output>Host 10.0.0.2 port 24007 already in peer"
                        " list</output>",
                        "</cliOutput>",
                        "",
                    ]
                ),
                "hostname": "\n".join(
                    [
                        '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                        "  <cliOutput>",
                        "  <opRet>0</opRet>",
                        "  <opErrno>2</opErrno>",
                        "  <opErrstr/>",
                        "  <output>Host server2 port 24007 already in peer"
                        " list</output>",
                        "</cliOutput>",
                        "",
                    ]
                ),
            }
            success_reverse_already_peer = {
                "ip": "\n".join(
                    [
                        '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                        "  <cliOutput>",
                        "  <opRet>0</opRet>",
                        "  <opErrno>2</opErrno>",
                        "  <opErrstr/>",
                        "  <output>Host 10.0.0.1 port 24007 already in peer"
                        " list</output>",
                        "</cliOutput>",
                        "",
                    ]
                ),
                "hostname": "\n".join(
                    [
                        '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                        "  <cliOutput>",
                        "  <opRet>0</opRet>",
                        "  <opErrno>2</opErrno>",
                        "  <opErrstr/>",
                        "  <output>Host server1 port 24007 already in peer"
                        " list</output>",
                        "</cliOutput>",
                        "",
                    ]
                ),
            }
            success_first_hostname_from_second_first_time = (
                success_reverse_already_peer["hostname"]
            )
            success_first_ip_from_second_first_time = success_other
            success_first_ip_from_second_second_time = success_reverse_already_peer[
                "ip"
            ]


#  gluster --version output collected in the wild.
version_output_362 = """
glusterfs 3.6.2 built on Jan 22 2015 12:59:57
Repository revision: git://git.gluster.com/glusterfs.git
Copyright (c) 2006-2011 Gluster Inc. <http://www.gluster.com>
GlusterFS comes with ABSOLUTELY NO WARRANTY.
You may redistribute copies of GlusterFS under the terms of the GNU General Public License
   """

version_output_61 = """
glusterfs 6.1
Repository revision: git://git.gluster.org/glusterfs.git
Copyright (c) 2006-2016 Red Hat, Inc. <https://www.gluster.org/>
GlusterFS comes with ABSOLUTELY NO WARRANTY.
It is licensed to you under your choice of the GNU Lesser
General Public License, version 3 or any later version (LGPLv3
or later), or the GNU General Public License, version 2 (GPLv2),
in all cases as published by the Free Software Foundation.
"""

xml_peer_present = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
  <peer>
    <uuid>uuid1</uuid>
    <hostname>node02</hostname>
    <hostnames>
      <hostname>node02.domain.dom</hostname>
      <hostname>10.0.0.2</hostname>
    </hostnames>
  </peer>
</cliOutput>
"""

xml_volume_present = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
  <volList>
    <volume>Newvolume1</volume>
    <volume>Newvolume2</volume>
  </volList>
</cliOutput>
"""

xml_volume_absent = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
  <volList>
    <count>0</count>
  </volList>
</cliOutput>
"""

xml_volume_status = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
  <volStatus>
    <volumes>
      <volume>
        <volName>myvol1</volName>
        <nodeCount>3</nodeCount>
        <node>
          <hostname>node01</hostname>
          <path>/tmp/foo</path>
          <peerid>830700d7-0684-497c-a12c-c02e365fb90b</peerid>
          <status>1</status>
          <port>49155</port>
          <ports>
            <tcp>49155</tcp>
            <rdma>N/A</rdma>
          </ports>
          <pid>2470</pid>
        </node>
        <node>
          <hostname>NFS Server</hostname>
          <path>localhost</path>
          <peerid>830700d7-0684-497c-a12c-c02e365fb90b</peerid>
          <status>0</status>
          <port>N/A</port>
          <ports>
            <tcp>N/A</tcp>
            <rdma>N/A</rdma>
          </ports>
          <pid>-1</pid>
        </node>
        <tasks/>
      </volume>
    </volumes>
  </volStatus>
</cliOutput>
"""

xml_volume_info_running = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
  <volInfo>
    <volumes>
      <volume>
        <name>myvol1</name>
        <id>f03c2180-cf55-4f77-ae0b-3650f57c82a1</id>
        <status>1</status>
        <statusStr>Started</statusStr>
        <brickCount>1</brickCount>
        <distCount>1</distCount>
        <stripeCount>1</stripeCount>
        <replicaCount>1</replicaCount>
        <disperseCount>0</disperseCount>
        <redundancyCount>0</redundancyCount>
        <type>0</type>
        <typeStr>Distribute</typeStr>
        <transport>0</transport>
        <bricks>
          <brick uuid="830700d7-0684-497c-a12c-c02e365fb90b">node01:/tmp/foo<name>node01:/tmp/foo</name><hostUuid>830700d7-0684-497c-a12c-c02e365fb90b</hostUuid></brick>
        </bricks>
        <optCount>1</optCount>
        <options>
          <option>
            <name>performance.readdir-ahead</name>
            <value>on</value>
          </option>
        </options>
      </volume>
      <count>1</count>
    </volumes>
  </volInfo>
</cliOutput>
"""

xml_volume_info_stopped = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
  <volInfo>
    <volumes>
      <volume>
        <name>myvol1</name>
        <status>1</status>
      </volume>
    </volumes>
  </volInfo>
</cliOutput>
"""

xml_peer_probe_success = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
  <opErrno>0</opErrno>
  <opErrstr/>
  <output/>
</cliOutput>
"""

xml_peer_probe_already_member = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
  <opErrno>2</opErrno>
  <opErrstr/>
  <output>Host salt port 24007 already in peer list</output>
</cliOutput>
"""

xml_peer_probe_localhost = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
  <opErrno>1</opErrno>
  <opErrstr/>
  <output>Probe on localhost not needed</output>
</cliOutput>
"""

xml_peer_probe_fail_cant_connect = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>-1</opRet>
  <opErrno>107</opErrno>
  <opErrstr>Probe returned with Transport endpoint is not connected</opErrstr>
</cliOutput>
"""

xml_command_success = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
</cliOutput>
"""

xml_command_fail = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>-1</opRet>
    <opErrno>0</opErrno>
  <opErrstr>Command Failed</opErrstr>
</cliOutput>
"""

xml_op_version_37 = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
  <opErrno>0</opErrno>
  <opErrstr/>
  <volGetopts>
    <count>1</count>
    <Option>cluster.op-version</Option>
    <Value>30707</Value>
  </volGetopts>
</cliOutput>
"""

xml_op_version_312 = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
  <opErrno>0</opErrno>
  <opErrstr/>
  <volGetopts>
    <count>1</count>
    <Opt>
      <Option>cluster.op-version</Option>
      <Value>30707</Value>
    </Opt>
  </volGetopts>
</cliOutput>
"""

xml_max_op_version = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
  <opErrno>0</opErrno>
  <opErrstr/>
  <volGetopts>
    <count>1</count>
    <Opt>
      <Option>cluster.max-op-version</Option>
      <Value>31200</Value>
    </Opt>
  </volGetopts>
</cliOutput>
"""

xml_set_op_version_failure = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>-1</opRet>
  <opErrno>30800</opErrno>
  <opErrstr>Required op-version (30707) should not be equal or lower than current cluster op-version (30707).</opErrstr>
  <cliOp>volSet</cliOp>
  <output>Set volume unsuccessful</output>
</cliOutput>
"""

xml_set_op_version_success = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>0</opRet>
  <opErrno>0</opErrno>
  <opErrstr/>
  <cliOp>volSet</cliOp>
  <output>Set volume successful</output>
</cliOutput>
"""


@pytest.fixture
def configure_loader_modules():
    return {glusterfs: {}}


maxDiff = None

# 'peer_status' function tests: 1


def test__get_version():
    """
    Test parsing of gluster --version.
    """
    mock_version = MagicMock(return_value="foo")
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock_version}):
        assert glusterfs._get_version() == (3, 6), "default behaviour"

    mock_version = MagicMock(return_value=version_output_362)
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock_version}):
        assert glusterfs._get_version() == (3, 6, 2)

    mock_version = MagicMock(return_value=version_output_61)
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock_version}):
        assert glusterfs._get_version() == (6, 1)

    more_versions = {
        "6.0": (6, 0),
        "4.1.10": (4, 1, 10),
        "5.13": (5, 13),
        "10.0": (10, 0),
    }
    for v in more_versions:
        mock_version = MagicMock(return_value=f"glusterfs {v}")
        with patch.dict(glusterfs.__salt__, {"cmd.run": mock_version}):
            assert glusterfs._get_version() == more_versions[v]


def test_peer_status():
    """
    Test gluster peer status
    """
    mock_run = MagicMock(return_value=xml_peer_present)
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
        assert glusterfs.peer_status() == {
            "uuid1": {"hostnames": ["node02", "node02.domain.dom", "10.0.0.2"]}
        }

    mock_run = MagicMock(return_value=xml_command_success)
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
        assert glusterfs.peer_status() == {}


# 'peer' function tests: 1


def test_peer():
    """
    Test if gluster peer call is successful.
    """
    mock_run = MagicMock()
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
        mock_run.return_value = xml_peer_probe_already_member
        assert glusterfs.peer("salt")

        mock_run.return_value = xml_peer_probe_localhost
        assert glusterfs.peer("salt")

        mock_run.return_value = xml_peer_probe_fail_cant_connect
        assert not glusterfs.peer("salt")


# 'create_volume' function tests: 1


def test_create_volume():
    """
    Test if it creates a glusterfs volume.
    """
    mock_run = MagicMock(return_value=xml_command_success)
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
        pytest.raises(
            SaltInvocationError, glusterfs.create_volume, "newvolume", "host1:brick"
        )

        pytest.raises(
            SaltInvocationError, glusterfs.create_volume, "newvolume", "host1/brick"
        )

        assert not mock_run.called

        mock_start_volume = MagicMock(return_value=True)
        with patch.object(glusterfs, "start_volume", mock_start_volume):
            # Create, do not start
            assert glusterfs.create_volume("newvolume", "host1:/brick")
            assert not mock_start_volume.called

            # Create and start
            assert glusterfs.create_volume("newvolume", "host1:/brick", start=True)
            assert mock_start_volume.called

            mock_start_volume.return_value = False
            # Create and fail start
            assert not glusterfs.create_volume("newvolume", "host1:/brick", start=True)

        mock_run.return_value = xml_command_fail
        assert not glusterfs.create_volume(
            "newvolume", "host1:/brick", True, True, True, "tcp", True
        )


# 'list_volumes' function tests: 1


def test_list_volumes():
    """
    Test if it list configured volumes
    """
    mock = MagicMock(return_value=xml_volume_absent)
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock}):
        assert glusterfs.list_volumes() == []

    mock = MagicMock(return_value=xml_volume_present)
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock}):
        assert glusterfs.list_volumes() == ["Newvolume1", "Newvolume2"]


# 'status' function tests: 1


def test_status():
    """
    Test if it check the status of a gluster volume.
    """
    mock_run = MagicMock(return_value=xml_command_fail)
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
        assert glusterfs.status("myvol1") is None

    res = {
        "bricks": {
            "node01:/tmp/foo": {
                "host": "node01",
                "hostname": "node01",
                "online": True,
                "path": "/tmp/foo",
                "peerid": "830700d7-0684-497c-a12c-c02e365fb90b",
                "pid": "2470",
                "port": "49155",
                "ports": {"rdma": "N/A", "tcp": "49155"},
                "status": "1",
            }
        },
        "healers": {},
        "nfs": {
            "node01": {
                "host": "NFS Server",
                "hostname": "NFS Server",
                "online": False,
                "path": "localhost",
                "peerid": "830700d7-0684-497c-a12c-c02e365fb90b",
                "pid": "-1",
                "port": "N/A",
                "ports": {"rdma": "N/A", "tcp": "N/A"},
                "status": "0",
            }
        },
    }
    mock = MagicMock(return_value=xml_volume_status)
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock}):
        assert glusterfs.status("myvol1") == res


# 'start_volume' function tests: 1


def test_volume_info():
    """
    Test if it returns the volume info.
    """
    res = {
        "myvol1": {
            "brickCount": "1",
            "bricks": {
                "brick1": {
                    "hostUuid": "830700d7-0684-497c-a12c-c02e365fb90b",
                    "path": "node01:/tmp/foo",
                    "uuid": "830700d7-0684-497c-a12c-c02e365fb90b",
                }
            },
            "disperseCount": "0",
            "distCount": "1",
            "id": "f03c2180-cf55-4f77-ae0b-3650f57c82a1",
            "name": "myvol1",
            "optCount": "1",
            "options": {"performance.readdir-ahead": "on"},
            "redundancyCount": "0",
            "replicaCount": "1",
            "status": "1",
            "statusStr": "Started",
            "stripeCount": "1",
            "transport": "0",
            "type": "0",
            "typeStr": "Distribute",
        }
    }
    mock = MagicMock(return_value=xml_volume_info_running)
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock}):
        assert glusterfs.info("myvol1") == res


def test_start_volume():
    """
    Test if it start a gluster volume.
    """
    # Stopped volume
    mock_info = MagicMock(return_value={"Newvolume1": {"status": "0"}})
    with patch.object(glusterfs, "info", mock_info):
        mock_run = MagicMock(return_value=xml_command_success)
        with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
            assert glusterfs.start_volume("Newvolume1") is True
            assert glusterfs.start_volume("nonExisting") is False
        mock_run = MagicMock(return_value=xml_command_fail)
        with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
            assert glusterfs.start_volume("Newvolume1") is False

    # Started volume
    mock_info = MagicMock(return_value={"Newvolume1": {"status": "1"}})
    with patch.object(glusterfs, "info", mock_info):
        mock_run = MagicMock(return_value=xml_command_success)
        with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
            assert glusterfs.start_volume("Newvolume1", force=True) is True
        mock_run = MagicMock(return_value=xml_command_fail)
        with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
            # cmd.run should not be called for already running volume:
            assert glusterfs.start_volume("Newvolume1") is True
            # except when forcing:
            assert glusterfs.start_volume("Newvolume1", force=True) is False


# 'stop_volume' function tests: 1


def test_stop_volume():
    """
    Test if it stop a gluster volume.
    """
    # Stopped volume
    mock_info = MagicMock(return_value={"Newvolume1": {"status": "0"}})
    with patch.object(glusterfs, "info", mock_info):
        mock_run = MagicMock(return_value=xml_command_success)
        with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
            assert glusterfs.stop_volume("Newvolume1") is True
            assert glusterfs.stop_volume("nonExisting") is False
        mock_run = MagicMock(return_value=xml_command_fail)
        with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
            # cmd.run should not be called for already stopped volume:
            assert glusterfs.stop_volume("Newvolume1") is True

    # Started volume
    mock_info = MagicMock(return_value={"Newvolume1": {"status": "1"}})
    with patch.object(glusterfs, "info", mock_info):
        mock_run = MagicMock(return_value=xml_command_success)
        with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
            assert glusterfs.stop_volume("Newvolume1") is True
            assert glusterfs.stop_volume("nonExisting") is False
        mock_run = MagicMock(return_value=xml_command_fail)
        with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
            assert glusterfs.stop_volume("Newvolume1") is False


# 'delete_volume' function tests: 1


def test_delete_volume():
    """
    Test if it deletes a gluster volume.
    """
    mock_info = MagicMock(return_value={"Newvolume1": {"status": "1"}})
    with patch.object(glusterfs, "info", mock_info):
        # volume doesn't exist
        assert not glusterfs.delete_volume("Newvolume3")

        mock_stop_volume = MagicMock(return_value=True)
        mock_run = MagicMock(return_value=xml_command_success)
        with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
            with patch.object(glusterfs, "stop_volume", mock_stop_volume):
                # volume exists, should not be stopped, and is started
                assert not glusterfs.delete_volume("Newvolume1", False)
                assert not mock_run.called
                assert not mock_stop_volume.called

                # volume exists, should be stopped, and is started
                assert glusterfs.delete_volume("Newvolume1")
                assert mock_run.called
                assert mock_stop_volume.called

    # volume exists and isn't started
    mock_info = MagicMock(return_value={"Newvolume1": {"status": "2"}})
    with patch.object(glusterfs, "info", mock_info):
        mock_run = MagicMock(return_value=xml_command_success)
        with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
            assert glusterfs.delete_volume("Newvolume1")
            mock_run.return_value = xml_command_fail
            assert not glusterfs.delete_volume("Newvolume1")


# 'add_volume_bricks' function tests: 1


def test_add_volume_bricks():
    """
    Test if it add brick(s) to an existing volume
    """
    mock_info = MagicMock(
        return_value={
            "Newvolume1": {
                "status": "1",
                "bricks": {
                    "brick1": {"path": "host:/path1"},
                    "brick2": {"path": "host:/path2"},
                },
            }
        }
    )
    with patch.object(glusterfs, "info", mock_info):
        mock_run = MagicMock(return_value=xml_command_success)
        with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
            # Volume does not exist
            assert not glusterfs.add_volume_bricks("nonExisting", ["bricks"])
            # Brick already exists
            assert glusterfs.add_volume_bricks("Newvolume1", ["host:/path2"])
            # Already existing brick as a string
            assert glusterfs.add_volume_bricks("Newvolume1", "host:/path2")
            assert not mock_run.called
            # A new brick:
            assert glusterfs.add_volume_bricks("Newvolume1", ["host:/new1"])
            assert mock_run.called

            # Gluster call fails
            mock_run.return_value = xml_command_fail
            assert not glusterfs.add_volume_bricks("Newvolume1", ["new:/path"])


# 'get_op_version' function tests: 1


def test_get_op_version():
    """
    Test retrieving the glusterfs op-version
    """

    # Test with xml output structure from v3.7
    mock_run = MagicMock(return_value=xml_op_version_37)
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
        assert glusterfs.get_op_version("test") == "30707"

    # Test with xml output structure from v3.12
    mock_run = MagicMock(return_value=xml_op_version_312)
    with patch.dict(glusterfs.__salt__, {"cmd.run": mock_run}):
        assert glusterfs.get_op_version("test") == "30707"


# 'get_max_op_version' function tests: 1


def test_get_max_op_version():
    """
    Test retrieving the glusterfs max-op-version.
    """

    mock_xml = MagicMock(return_value=xml_max_op_version)
    mock_version = MagicMock(return_value="glusterfs 3.9.1")

    with patch.dict(glusterfs.__salt__, {"cmd.run": mock_version}):
        assert not glusterfs.get_max_op_version()[0]

    with patch.object(glusterfs, "_get_version", return_value=(3, 12, 0)):
        with patch.dict(glusterfs.__salt__, {"cmd.run": mock_xml}):
            assert glusterfs.get_max_op_version() == "31200"


# 'set_op_version' function tests: 1


def test_set_op_version():
    """
    Test setting the glusterfs op-version
    """
    mock_failure = MagicMock(return_value=xml_set_op_version_failure)
    mock_success = MagicMock(return_value=xml_set_op_version_success)

    with patch.dict(glusterfs.__salt__, {"cmd.run": mock_failure}):
        assert not glusterfs.set_op_version(30707)[0]

    with patch.dict(glusterfs.__salt__, {"cmd.run": mock_success}):
        assert glusterfs.set_op_version(31200) == "Set volume successful"



import copy
import logging
import os
import pathlib
import pprint
import re
import shutil
import sys

import pytest
from saltfactories.utils import random_string

import salt.defaults.exitcodes
import salt.utils.files
import salt.utils.json
import salt.utils.platform
import salt.utils.yaml
import tests.conftest
import tests.support.helpers
from tests.conftest import FIPS_TESTRUN
from tests.support.runtests import RUNTIME_VARS

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.windows_whitelisted,
]

log = logging.getLogger(__name__)


def test_fib(salt_call_cli):
    ret = salt_call_cli.run("test.fib", "3")
    assert ret.returncode == 0
    assert ret.data[0] == 2


def test_fib_txt_output(salt_call_cli):
    ret = salt_call_cli.run("--output=txt", "test.fib", "3")
    assert ret.returncode == 0
    assert ret.data is None
    assert (
        re.match(r"local: \(2, [0-9]{1}\.(([0-9]+)(e-([0-9]+))?)\)\s", ret.stdout)
        is not None
    )


@pytest.mark.parametrize("indent", [-1, 0, 1])
def test_json_out_indent(salt_call_cli, indent):
    ret = salt_call_cli.run("--out=json", f"--out-indent={indent}", "test.ping")
    assert ret.returncode == 0
    assert ret.data is True
    if indent == -1:
        expected_output = '{"local": true}\n'
    elif indent == 0:
        expected_output = '{\n"local": true\n}\n'
    else:
        expected_output = '{\n "local": true\n}\n'
    stdout = ret.stdout
    assert ret.stdout == expected_output


def test_local_sls_call(salt_master, salt_call_cli):
    sls_contents = """
    regular-module:
      module.run:
        - name: test.echo
        - text: hello
    """
    with salt_master.state_tree.base.temp_file("saltcalllocal.sls", sls_contents):
        ret = salt_call_cli.run(
            "--local",
            "--file-root",
            str(salt_master.state_tree.base.paths[0]),
            "state.sls",
            "saltcalllocal",
        )
        assert ret.returncode == 0
        state_run_dict = next(iter(ret.data.values()))
        assert state_run_dict["name"] == "test.echo"
        assert state_run_dict["result"] is True
        assert state_run_dict["changes"]["ret"] == "hello"


def test_local_salt_call(salt_call_cli):
    """
    This tests to make sure that salt-call does not execute the
    function twice, see https://github.com/saltstack/salt/pull/49552
    """
    with pytest.helpers.temp_file() as filename:

        ret = salt_call_cli.run(
            "--local", "state.single", "file.append", name=str(filename), text="foo"
        )
        assert ret.returncode == 0

        state_run_dict = next(iter(ret.data.values()))
        assert state_run_dict["changes"]

        # 2nd sanity check: make sure that "foo" only exists once in the file
        contents = filename.read_text()
        assert contents.count("foo") == 1, contents


@pytest.mark.skip_on_windows(reason=tests.support.helpers.PRE_PYTEST_SKIP_REASON)
def test_user_delete_kw_output(salt_call_cli):
    ret = salt_call_cli.run("-d", "user.delete", _timeout=120)
    assert ret.returncode == 0
    expected_output = "salt '*' user.delete name"
    if not salt.utils.platform.is_windows():
        expected_output += " remove=True force=True"
    assert expected_output in ret.stdout


def test_salt_documentation_too_many_arguments(salt_call_cli):
    """
    Test to see if passing additional arguments shows an error
    """
    ret = salt_call_cli.run("-d", "virtualenv.create", "/tmp/ve")
    assert ret.returncode != 0
    assert "You can only get documentation for one method at one time" in ret.stderr


def test_issue_6973_state_highstate_exit_code(salt_call_cli):
    """
    If there is no tops/master_tops or state file matches
    for this minion, salt-call should exit non-zero if invoked with
    option --retcode-passthrough
    """
    expected_comment = "No states found for this minion"
    ret = salt_call_cli.run("--retcode-passthrough", "state.highstate")
    assert ret.returncode != 0
    assert expected_comment in ret.stdout


@tests.support.helpers.PRE_PYTEST_SKIP
def test_issue_15074_output_file_append(salt_call_cli):

    with pytest.helpers.temp_file(name="issue-15074") as output_file_append:
        ret = salt_call_cli.run(
            "--output-file", str(output_file_append), "test.versions"
        )
        assert ret.returncode == 0

        first_run_output = output_file_append.read_text()

        assert first_run_output

        ret = salt_call_cli.run(
            "--output-file",
            str(output_file_append),
            "--output-file-append",
            "test.versions",
        )
        assert ret.returncode == 0

        second_run_output = output_file_append.read_text()

        assert second_run_output

        assert second_run_output == first_run_output + first_run_output


@tests.support.helpers.PRE_PYTEST_SKIP
def test_issue_14979_output_file_permissions(salt_call_cli):
    with pytest.helpers.temp_file(name="issue-14979") as output_file:
        with salt.utils.files.set_umask(0o077):
            # Let's create an initial output file with some data
            ret = salt_call_cli.run("--output-file", str(output_file), "--grains")
            assert ret.returncode == 0
            try:
                stat1 = output_file.stat()
            except OSError:
                pytest.fail(f"Failed to generate output file {output_file}")

            # Let's change umask
            os.umask(0o777)  # pylint: disable=blacklisted-function

            ret = salt_call_cli.run(
                "--output-file", str(output_file), "--output-file-append", "--grains"
            )
            assert ret.returncode == 0
            stat2 = output_file.stat()
            assert stat1.st_mode == stat2.st_mode
            # Data was appeneded to file
            assert stat1.st_size < stat2.st_size

            # Let's remove the output file
            output_file.unlink()

            # Not appending data
            ret = salt_call_cli.run("--output-file", str(output_file), "--grains")
            assert ret.returncode == 0
            try:
                stat3 = output_file.stat()
            except OSError:
                pytest.fail(f"Failed to generate output file {output_file}")
            # Mode must have changed since we're creating a new log file
            assert stat1.st_mode != stat3.st_mode


@pytest.mark.skip_on_windows(reason="This test does not apply on Win")
def test_42116_cli_pillar_override(salt_call_cli):
    ret = salt_call_cli.run(
        "state.apply",
        "issue-42116-cli-pillar-override",
        pillar={"myhost": "localhost"},
    )
    state_run_dict = next(iter(ret.data.values()))
    assert state_run_dict["changes"]
    assert (
        state_run_dict["comment"] == 'Command "ping -c 2 localhost" run'
    ), "CLI pillar override not found in pillar data. State Run Dictionary:\n{}".format(
        pprint.pformat(state_run_dict)
    )


def test_pillar_items_masterless(salt_minion, salt_call_cli):
    """
    Test to ensure we get expected output
    from pillar.items with salt-call
    """
    top_file = """
    base:
      '{}':
        - basic
    """.format(
        salt_minion.id
    )
    basic_pillar_file = """
    monty: python
    knights:
      - Lancelot
      - Galahad
      - Bedevere
      - Robin
    """
    top_tempfile = salt_minion.pillar_tree.base.temp_file("top.sls", top_file)
    basic_tempfile = salt_minion.pillar_tree.base.temp_file(
        "basic.sls", basic_pillar_file
    )

    with top_tempfile, basic_tempfile:
        ret = salt_call_cli.run("--local", "pillar.items")
        assert ret.returncode == 0
        assert "knights" in ret.data
        assert sorted(ret.data["knights"]) == sorted(
            ["Lancelot", "Galahad", "Bedevere", "Robin"]
        )
        assert "monty" in ret.data
        assert ret.data["monty"] == "python"


def test_masterless_highstate(salt_minion, salt_call_cli, tmp_path):
    """
    test state.highstate in masterless mode
    """
    top_sls = """
    base:
      '*':
        - core
        """

    testfile = tmp_path / "testfile"
    core_state = """
    {}:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
        """.format(
        testfile
    )

    expected_id = str(testfile)

    with salt_minion.state_tree.base.temp_file(
        "top.sls", top_sls
    ), salt_minion.state_tree.base.temp_file("core.sls", core_state):
        ret = salt_call_cli.run("--local", "state.highstate")
        assert ret.returncode == 0
        state_run_dict = next(iter(ret.data.values()))
        assert state_run_dict["result"] is True
        assert state_run_dict["__id__"] == expected_id


@pytest.mark.skip_on_windows
def test_syslog_file_not_found(salt_minion, salt_call_cli, tmp_path):
    """
    test when log_file is set to a syslog file that does not exist
    """
    config_dir = tmp_path / "log_file_incorrect"
    config_dir.mkdir()
    with pytest.helpers.change_cwd(str(config_dir)):
        minion_config = copy.deepcopy(salt_minion.config)
        minion_config["log_file"] = "file:///dev/doesnotexist"
        with salt.utils.files.fopen(str(config_dir / "minion"), "w") as fh_:
            fh_.write(salt.utils.yaml.dump(minion_config, default_flow_style=False))
        ret = salt_call_cli.run(
            f"--config-dir={config_dir}",
            "--log-level=debug",
            "cmd.run",
            "echo foo",
        )
        if sys.version_info >= (3, 5, 4):
            assert ret.returncode == 0
            assert (
                "[WARNING ] The log_file does not exist. Logging not setup correctly or"
                " syslog service not started." in ret.stderr
            )
            assert ret.data == "foo", ret
        else:
            assert ret.returncode == salt.defaults.exitcodes.EX_UNAVAILABLE
            assert "Failed to setup the Syslog logging handler" in ret.stderr


@tests.support.helpers.PRE_PYTEST_SKIP
@pytest.mark.skip_on_windows
def test_return(salt_call_cli, salt_run_cli):
    command = "echo returnTOmaster"
    ret = salt_call_cli.run("cmd.run", command)
    assert ret.returncode == 0
    assert ret.data == "returnTOmaster"

    ret = salt_run_cli.run("jobs.list_jobs")
    assert ret.returncode == 0
    jid = target = None
    for jid, details in ret.data.items():
        if command in details["Arguments"]:
            target = details["Target"]
            break

    ret = salt_run_cli.run("jobs.lookup_jid", jid, _timeout=60)
    assert ret.returncode == 0
    assert target in ret.data
    assert ret.data[target] == "returnTOmaster"


def test_exit_status_unknown_argument(salt_call_cli):
    """
    Ensure correct exit status when an unknown argument is passed to salt CLI.
    """
    ret = salt_call_cli.run("--unknown-argument")
    assert ret.returncode == salt.defaults.exitcodes.EX_USAGE, ret
    assert "Usage" in ret.stderr
    assert "no such option: --unknown-argument" in ret.stderr


def test_exit_status_correct_usage(salt_call_cli):
    """
    Ensure correct exit status when salt CLI starts correctly.

    """
    ret = salt_call_cli.run("test.true")
    assert ret.returncode == salt.defaults.exitcodes.EX_OK, ret


def test_context_retcode_salt_call(salt_call_cli):
    """
    Test that a nonzero retcode set in the context dunder will cause the
    salt CLI to set a nonzero retcode.
    """
    # Test salt-call, making sure to also confirm the behavior of
    # retcode_passthrough.
    ret = salt_call_cli.run("test.retcode", "0")
    assert ret.returncode == 0, ret
    ret = salt_call_cli.run("test.retcode", "42")
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret
    ret = salt_call_cli.run("--retcode-passthrough", "test.retcode", "42")
    assert ret.returncode == 42, ret

    # Test a state run that exits with one or more failures
    ret = salt_call_cli.run("state.single", "test.fail_without_changes", "foo")
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret
    ret = salt_call_cli.run(
        "--retcode-passthrough", "state.single", "test.fail_without_changes", "foo"
    )
    assert ret.returncode == salt.defaults.exitcodes.EX_STATE_FAILURE, ret

    # Test a state compiler error
    ret = salt_call_cli.run("state.apply", "thisslsfiledoesnotexist")
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret
    ret = salt_call_cli.run(
        "--retcode-passthrough", "state.apply", "thisslsfiledoesnotexist"
    )
    assert ret.returncode == salt.defaults.exitcodes.EX_STATE_COMPILER_ERROR, ret


def test_salt_call_error(salt_call_cli):
    """
    Test that we return the expected retcode when a minion function raises
    an exception.
    """
    ret = salt_call_cli.run("test.raise_exception", "TypeError")
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret

    ret = salt_call_cli.run(
        "test.raise_exception", "salt.exceptions.CommandNotFoundError"
    )
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret

    ret = salt_call_cli.run(
        "test.raise_exception", "salt.exceptions.CommandExecutionError"
    )
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret

    ret = salt_call_cli.run(
        "test.raise_exception", "salt.exceptions.SaltInvocationError"
    )
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret

    ret = salt_call_cli.run(
        "test.raise_exception",
        "OSError",
        "2",
        "No such file or directory",
        "/tmp/foo.txt",
    )
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret

    ret = salt_call_cli.run("test.echo", "{foo: bar, result: False}")
    assert ret.returncode == salt.defaults.exitcodes.EX_GENERIC, ret


def test_local_salt_call_no_function_no_retcode(salt_call_cli):
    """
    This tests ensures that when salt-call --local is called
    with a module but without a function the return code is 1
    and we receive the docs for all module functions.

    Also ensure we don't get an exception.
    """
    ret = salt_call_cli.run("--local", "test")
    assert ret.returncode == 1
    assert ret.data
    assert "test" in ret.data
    assert ret.data["test"] == "'test' is not available."
    assert "test.echo" in ret.data


@pytest.fixture
def master_id_alt():
    master_id = random_string("master-")
    yield master_id


@pytest.fixture
def minion_id_alt():
    master_id = random_string("minion-")
    yield master_id


@pytest.fixture
def salt_master_alt(salt_factories, tmp_path, master_id_alt):
    """
    A running salt-master fixture
    """
    root_dir = salt_factories.get_root_dir_for_daemon(master_id_alt)
    conf_dir = root_dir / "conf"
    conf_dir.mkdir(exist_ok=True)
    extension_modules_path = str(root_dir / "extension_modules")
    if not os.path.exists(extension_modules_path):
        shutil.copytree(
            os.path.join(RUNTIME_VARS.FILES, "extension_modules"),
            extension_modules_path,
        )
    cache = pathlib.Path(extension_modules_path) / "cache"
    cache.mkdir()
    localfs = cache / "localfs.py"
    localfs.write_text(
        tests.support.helpers.dedent(
            """
        from salt.exceptions import SaltClientError
        def store(bank, key, data): # , cachedir):
            raise SaltClientError("TEST")
        """
        )
    )
    factory = salt_factories.salt_master_daemon(
        master_id_alt,
        defaults={
            "root_dir": str(root_dir),
            "extension_modules": extension_modules_path,
            "auto_accept": True,
        },
        overrides={
            "fips_mode": FIPS_TESTRUN,
            "publish_signing_algorithm": (
                "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
            ),
        },
    )
    with factory.pillar_tree.base.temp_file("cve_2024_37088.sls", "foobar: bang"):
        with factory.state_tree.base.temp_file(
            "cve_2024_37088.sls",
            """
            # cvs_2024_37088.sls
            {{%- set var = salt ['pillar.get']('foobar', 'state default') %}}

            test:
              file.managed:
                - name: {0}
                - contents: {{{{ var }}}}
            """.format(
                tmp_path / "cve_2024_37088.txt"
            ),
        ):
            with factory.started():
                yield factory


@pytest.fixture
def salt_call_alt(salt_master_alt, minion_id_alt):
    minion_factory = salt_master_alt.salt_minion_daemon(
        minion_id_alt,
        overrides={
            "fips_mode": tests.conftest.FIPS_TESTRUN,
            "encryption_algorithm": (
                "OAEP-SHA224" if tests.conftest.FIPS_TESTRUN else "OAEP-SHA1"
            ),
            "signing_algorithm": (
                "PKCS1v15-SHA224" if tests.conftest.FIPS_TESTRUN else "PKCS1v15-SHA1"
            ),
        },
    )
    return minion_factory.salt_call_cli()


def test_cve_2024_37088(salt_master_alt, salt_call_alt, caplog):
    with caplog.at_level(logging.ERROR):
        ret = salt_call_alt.run("state.sls", "cve_2024_37088")
        assert ret.returncode == 1
        assert ret.data is None
        assert "Got a bad pillar from master, type str, expecting dict" in caplog.text



"""
unit tests for the jobs runner
"""

import pytest

import salt.minion
import salt.runners.jobs as jobs
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {
        jobs: {"__opts__": {"ext_job_cache": None, "master_job_cache": "local_cache"}}
    }


def test_list_jobs_with_search_target():
    """
    test jobs.list_jobs runner with search_target args
    """
    mock_jobs_cache = {
        "20160524035503086853": {
            "Arguments": [],
            "Function": "test.ping",
            "StartTime": "2016, May 24 03:55:03.086853",
            "Target": "node-1-1.com",
            "Target-type": "glob",
            "User": "root",
        },
        "20160524035524895387": {
            "Arguments": [],
            "Function": "test.ping",
            "StartTime": "2016, May 24 03:55:24.895387",
            "Target": ["node-1-2.com", "node-1-1.com"],
            "Target-type": "list",
            "User": "sudo_ubuntu",
        },
    }

    def return_mock_jobs():
        return mock_jobs_cache

    class MockMasterMinion:

        returners = {"local_cache.get_jids": return_mock_jobs}

        def __init__(self, *args, **kwargs):
            pass

    returns = {
        "all": mock_jobs_cache,
        "node-1-1.com": mock_jobs_cache,
        "node-1-2.com": {
            "20160524035524895387": mock_jobs_cache["20160524035524895387"]
        },
        "non-existant": {},
    }

    with patch.object(salt.minion, "MasterMinion", MockMasterMinion):
        assert jobs.list_jobs() == returns["all"]

        assert (
            jobs.list_jobs(search_target=["node-1-1*", "node-1-2*"]) == returns["all"]
        )

        assert jobs.list_jobs(search_target="node-1-1.com") == returns["node-1-1.com"]

        assert jobs.list_jobs(search_target="node-1-2.com") == returns["node-1-2.com"]

        assert jobs.list_jobs(search_target="non-existant") == returns["non-existant"]



def test_pillar_using_cp_module(salt_master, salt_minion, salt_cli, tmp_path):
    pillar_top = """
    base:
      "*":
        - my_pillar
    """
    my_file = tmp_path / "my_file.txt"
    my_file.write_text("foobar")
    my_pillar = f"""
    {{%- set something = salt['cp.get_file_str']("{str(my_file)}") %}}
    file_content: {{{{ something }}}}
    """
    with salt_master.pillar_tree.base.temp_file("top.sls", pillar_top):
        with salt_master.pillar_tree.base.temp_file("my_pillar.sls", my_pillar):

            ret = salt_cli.run("saltutil.pillar_refresh", minion_tgt=salt_minion.id)
            assert ret.returncode == 0

            pillar_ret = salt_cli.run(
                "pillar.item", "file_content", minion_tgt=salt_minion.id
            )
            assert pillar_ret.returncode == 0
            assert '"file_content": "foobar"' in pillar_ret.stdout



"""
Tests for the MySQL states
"""

import logging

import pytest

import salt.utils.path
from salt.modules import mysql as mysqlmod
from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin

log = logging.getLogger(__name__)

NO_MYSQL = False
try:
    import MySQLdb  # pylint: disable=import-error,unused-import
except ImportError:
    NO_MYSQL = True

if not salt.utils.path.which("mysqladmin"):
    NO_MYSQL = True


@pytest.mark.skipif(
    NO_MYSQL,
    reason="Please install MySQL bindings and a MySQL Server before running "
    "MySQL integration tests.",
)
class MysqlGrantsStateTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the mysql_grants states
    """

    user = "root"
    password = "poney"
    # yep, theses are valid MySQL db names
    # very special chars are _ % and .
    testdb1 = "tes.t'\"saltdb"
    testdb2 = "t_st `(:=salt%b)"
    testdb3 = "test `(:=salteeb)"
    table1 = "foo"
    table2 = "foo `'%_bar"
    users = {
        "user1": {"name": "foo", "pwd": "bar"},
        "user2": {"name": 'user ";--,?:&/\\', "pwd": '";--(),?:@=&/\\'},
        # this is : passwd 標標
        "user3": {"name": "user( @ )=foobar", "pwd": "\xe6\xa8\x99\xe6\xa8\x99"},
        # this is : user/password containing 標標
        "user4": {"name": "user \xe6\xa8\x99", "pwd": "\xe6\xa8\x99\xe6\xa8\x99"},
    }

    @pytest.mark.destructive_test
    def setUp(self):
        """
        Test presence of MySQL server, enforce a root password
        """
        super().setUp()
        NO_MYSQL_SERVER = True
        # now ensure we know the mysql root password
        # one of theses two at least should work
        ret1 = self.run_state(
            "cmd.run",
            name='mysqladmin --host="localhost" -u '
            + self.user
            + ' flush-privileges password "'
            + self.password
            + '"',
        )
        ret2 = self.run_state(
            "cmd.run",
            name='mysqladmin --host="localhost" -u '
            + self.user
            + ' --password="'
            + self.password
            + '" flush-privileges password "'
            + self.password
            + '"',
        )
        key, value = ret2.popitem()
        if value["result"]:
            NO_MYSQL_SERVER = False
        else:
            self.skipTest("No MySQL Server running, or no root access on it.")
        # Create some users and a test db
        for user, userdef in self.users.items():
            self._userCreation(uname=userdef["name"], password=userdef["pwd"])
        self.run_state(
            "mysql_database.present",
            name=self.testdb1,
            character_set="utf8",
            collate="utf8_general_ci",
            connection_user=self.user,
            connection_pass=self.password,
        )
        self.run_state(
            "mysql_database.present",
            name=self.testdb2,
            character_set="utf8",
            collate="utf8_general_ci",
            connection_user=self.user,
            connection_pass=self.password,
        )
        create_query = (
            "CREATE TABLE {tblname} ("
            " id INT NOT NULL AUTO_INCREMENT PRIMARY KEY,"
            " data VARCHAR(100)) ENGINE={engine};".format(
                tblname=mysqlmod.quote_identifier(self.table1),
                engine="MYISAM",
            )
        )
        log.info("Adding table '%s'", self.table1)
        self.run_function(
            "mysql.query",
            database=self.testdb2,
            query=create_query,
            connection_user=self.user,
            connection_pass=self.password,
        )
        create_query = (
            "CREATE TABLE {tblname} ("
            " id INT NOT NULL AUTO_INCREMENT PRIMARY KEY,"
            " data VARCHAR(100)) ENGINE={engine};".format(
                tblname=mysqlmod.quote_identifier(self.table2),
                engine="MYISAM",
            )
        )
        log.info("Adding table '%s'", self.table2)
        self.run_function(
            "mysql.query",
            database=self.testdb2,
            query=create_query,
            connection_user=self.user,
            connection_pass=self.password,
        )

    @pytest.mark.destructive_test
    def tearDown(self):
        """
        Removes created users and db
        """
        for user, userdef in self.users.items():
            self._userRemoval(uname=userdef["name"], password=userdef["pwd"])
        self.run_state(
            "mysql_database.absent",
            name=self.testdb1,
            connection_user=self.user,
            connection_pass=self.password,
        )
        self.run_function(
            "mysql_database.absent",
            name=self.testdb2,
            connection_user=self.user,
            connection_pass=self.password,
        )

    def _userCreation(self, uname, password=None):
        """
        Create a test user
        """
        self.run_state(
            "mysql_user.present",
            name=uname,
            host="localhost",
            password=password,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
            saltenv={"LC_ALL": "en_US.utf8"},
        )

    def _userRemoval(self, uname, password=None):
        """
        Removes a test user
        """
        self.run_state(
            "mysql_user.absent",
            name=uname,
            host="localhost",
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
            saltenv={"LC_ALL": "en_US.utf8"},
        )

    @pytest.mark.destructive_test
    def test_grant_present_absent(self):
        """
        mysql_database.present
        """
        ret = self.run_state(
            "mysql_grants.present",
            name="grant test 1",
            grant="SELECT, INSERT",
            database=self.testdb1 + ".*",
            user=self.users["user1"]["name"],
            host="localhost",
            grant_option=True,
            revoke_first=True,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "mysql_grants.present",
            name="grant test 2",
            grant="SELECT, ALTER,CREATE TEMPORARY tables, execute",
            database=self.testdb1 + ".*",
            user=self.users["user1"]["name"],
            host="localhost",
            grant_option=True,
            revoke_first=True,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "mysql_grants.present",
            name="grant test 3",
            grant="SELECT, INSERT",
            database=self.testdb2 + "." + self.table2,
            user=self.users["user2"]["name"],
            host="localhost",
            grant_option=True,
            revoke_first=True,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "mysql_grants.present",
            name="grant test 4",
            grant="SELECT, INSERT",
            database=self.testdb2 + "." + self.table2,
            user=self.users["user2"]["name"],
            host="localhost",
            grant_option=True,
            revoke_first=True,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "mysql_grants.present",
            name="grant test 5",
            grant="SELECT, UPDATE",
            database=self.testdb2 + ".*",
            user=self.users["user1"]["name"],
            host="localhost",
            grant_option=True,
            revoke_first=False,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "mysql_grants.absent",
            name="grant test 6",
            grant="SELECT,update",
            database=self.testdb2 + ".*",
            user=self.users["user1"]["name"],
            host="localhost",
            grant_option=True,
            revoke_first=False,
            connection_user=self.user,
            connection_pass=self.password,
            connection_charset="utf8",
        )
        self.assertSaltTrueReturn(ret)



"""
    :codeauthor: Logilab <contact@logilab.fr>
"""

import pytest

import salt.states.postgres_cluster as postgres_cluster
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {postgres_cluster: {}}


def test_present():
    """
    Test to ensure that the named database is present
    with the specified properties.
    """
    name = "main"
    version = "9.4"

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    mock_t = MagicMock(return_value=True)
    mock_f = MagicMock(return_value=False)
    infos = {f"{version}/{name}": {}}
    mock = MagicMock(return_value=infos)
    with patch.dict(
        postgres_cluster.__salt__,
        {
            "postgres.cluster_list": mock,
            "postgres.cluster_exists": mock_t,
            "postgres.cluster_create": mock_t,
        },
    ):
        comt = f"Cluster {version}/{name} is already present"
        ret.update({"comment": comt, "result": True})
        assert postgres_cluster.present(version, name) == ret
        infos[f"{version}/{name}"]["port"] = 5433
        comt = (
            "Cluster {}/{} has wrong parameters "
            "which couldn't be changed on fly.".format(version, name)
        )
        ret.update({"comment": comt, "result": False})
        assert postgres_cluster.present(version, name, port=5434) == ret
        infos[f"{version}/{name}"]["datadir"] = "/tmp/"
        comt = (
            "Cluster {}/{} has wrong parameters "
            "which couldn't be changed on fly.".format(version, name)
        )
        ret.update({"comment": comt, "result": False})
        assert postgres_cluster.present(version, name, port=5434) == ret

    with patch.dict(
        postgres_cluster.__salt__,
        {
            "postgres.cluster_list": mock,
            "postgres.cluster_exists": mock_f,
            "postgres.cluster_create": mock_t,
        },
    ):
        comt = f"The cluster {version}/{name} has been created"
        ret.update(
            {
                "comment": comt,
                "result": True,
                "changes": {f"{version}/{name}": "Present"},
            }
        )
        assert postgres_cluster.present(version, name) == ret
        with patch.dict(postgres_cluster.__opts__, {"test": True}):
            comt = f"Cluster {version}/{name} is set to be created"
            ret.update({"comment": comt, "result": None, "changes": {}})
            assert postgres_cluster.present(version, name) == ret

    with patch.dict(
        postgres_cluster.__salt__,
        {
            "postgres.cluster_list": mock,
            "postgres.cluster_exists": mock_f,
            "postgres.cluster_create": mock_f,
        },
    ):
        comt = f"Failed to create cluster {version}/{name}"
        ret.update({"comment": comt, "result": False})
        assert postgres_cluster.present(version, name) == ret


def test_absent():
    """
    Test to ensure that the named database is absent.
    """
    name = "main"
    version = "9.4"

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    mock_t = MagicMock(return_value=True)
    mock = MagicMock(side_effect=[True, True, False])
    with patch.dict(
        postgres_cluster.__salt__,
        {"postgres.cluster_exists": mock, "postgres.cluster_remove": mock_t},
    ):
        with patch.dict(postgres_cluster.__opts__, {"test": True}):
            comt = f"Cluster {version}/{name} is set to be removed"
            ret.update({"comment": comt, "result": None})
            assert postgres_cluster.absent(version, name) == ret

        with patch.dict(postgres_cluster.__opts__, {"test": False}):
            comt = f"Cluster {version}/{name} has been removed"
            ret.update({"comment": comt, "result": True, "changes": {name: "Absent"}})
            assert postgres_cluster.absent(version, name) == ret

            comt = "Cluster {}/{} is not present, so it cannot be removed".format(
                version, name
            )
            ret.update({"comment": comt, "result": True, "changes": {}})
            assert postgres_cluster.absent(version, name) == ret



import logging
import subprocess
import time

import pytest

log = logging.getLogger(__name__)


@pytest.mark.skip_unless_on_linux
def test_multimaster_dns(
    salt_mm_master_1,
    salt_mm_minion_1,
    mm_master_1_salt_cli,
    etc_hosts,
    caplog,
    master_alive_interval,
):
    """
    Verify a minion configured with multimaster hot/hot will pick up a master's
    dns change if it's been disconnected.
    """

    etc_hosts.write_text(
        f"{etc_hosts.orig_text}\n172.16.0.1    master1.local master2.local"
    )

    log.info("Added hosts record for master1.local and master2.local")

    with salt_mm_minion_1.started(start_timeout=180):
        with caplog.at_level(logging.INFO):
            ret = mm_master_1_salt_cli.run("test.ping", minion_tgt="mm-minion-1")
            assert ret.returncode == 0
            etc_hosts.write_text(
                f"{etc_hosts.orig_text}\n127.0.0.1    master1.local master2.local"
            )
            log.info("Changed hosts record for master1.local and master2.local")
            subprocess.check_output(["ip", "addr", "del", "172.16.0.1/32", "dev", "lo"])
            log.info("Removed secondary master IP address.")
            # Wait for the minion's master_alive_interval, adding a second for
            # reliablity.
            time.sleep(master_alive_interval + 10)
            assert (
                "Master ip address changed from 172.16.0.1 to 127.0.0.1" in caplog.text
            )
            ret = mm_master_1_salt_cli.run("test.ping", minion_tgt="mm-minion-1")
            assert ret.returncode == 0
            assert (
                "Master ip address changed from 172.16.0.1 to 127.0.0.1" in caplog.text
            )



"""
Manage JBoss 7 Application Server via CLI interface

.. versionadded:: 2015.5.0

This state uses the jboss-cli.sh script from a JBoss or Wildfly installation and parses its output to determine the execution result.

In order to run each state, a jboss_config dictionary with the following properties must be passed:

.. code-block:: yaml

   jboss:
      cli_path: '/opt/jboss/jboss-7.0/bin/jboss-cli.sh'
      controller: 10.11.12.13:9999
      cli_user: 'jbossadm'
      cli_password: 'jbossadm'

If the controller doesn't require a password, then the cli_user and cli_password parameters are optional.

Since same dictionary with configuration will be used in all the states, it may be more convenient to move JBoss configuration and other properties
to the pillar.

Example of application deployment from local filesystem:

.. code-block:: yaml

     application_deployed:
       jboss7.deployed:
         - salt_source:
             target_file: '/tmp/webapp.war'
         - jboss_config: {{ pillar['jboss'] }}

For the sake of brevity, examples for each state assume that jboss_config is contained in the pillar.


"""

import logging
import re
import time
import traceback

import salt.utils.dictdiffer as dictdiffer
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


def datasource_exists(
    name, jboss_config, datasource_properties, recreate=False, profile=None
):
    """
    Ensures that a datasource with given properties exist on the jboss instance.
    If datasource doesn't exist, it is created, otherwise only the properties that are different will be updated.

    name
        Datasource property name
    jboss_config
        Dict with connection properties (see state description)
    datasource_properties
        Dict with datasource properties
    recreate : False
        If set to True and datasource exists it will be removed and created again. However, if there are deployments that depend on the datasource, it will not me possible to remove it.
    profile : None
        The profile name for this datasource (domain mode only)

    Example:

    .. code-block:: yaml

        sampleDS:
          jboss7.datasource_exists:
           - recreate: False
           - datasource_properties:
               driver-name: mysql
               connection-url: 'jdbc:mysql://localhost:3306/sampleDatabase'
               jndi-name: 'java:jboss/datasources/sampleDS'
               user-name: sampleuser
               password: secret
               min-pool-size: 3
               use-java-context: True
           - jboss_config: {{ pillar['jboss'] }}
           - profile: full-ha

    """
    log.debug(
        " ======================== STATE: jboss7.datasource_exists (name: %s) ", name
    )
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    has_changed = False
    ds_current_properties = {}
    ds_result = __salt__["jboss7.read_datasource"](
        jboss_config=jboss_config, name=name, profile=profile
    )
    if ds_result["success"]:
        ds_current_properties = ds_result["result"]
        if recreate:
            remove_result = __salt__["jboss7.remove_datasource"](
                jboss_config=jboss_config, name=name, profile=profile
            )
            if remove_result["success"]:
                ret["changes"]["removed"] = name
            else:
                ret["result"] = False
                ret["comment"] = (
                    "Could not remove datasource. Stdout: " + remove_result["stdout"]
                )
                return ret

            has_changed = True  # if we are here, we have already made a change

            create_result = __salt__["jboss7.create_datasource"](
                jboss_config=jboss_config,
                name=name,
                datasource_properties=datasource_properties,
                profile=profile,
            )
            if create_result["success"]:
                ret["changes"]["created"] = name
            else:
                ret["result"] = False
                ret["comment"] = (
                    "Could not create datasource. Stdout: " + create_result["stdout"]
                )
                return ret

            read_result = __salt__["jboss7.read_datasource"](
                jboss_config=jboss_config, name=name, profile=profile
            )
            if read_result["success"]:
                ds_new_properties = read_result["result"]
            else:
                ret["result"] = False
                ret["comment"] = (
                    "Could not read datasource. Stdout: " + read_result["stdout"]
                )
                return ret

        else:
            update_result = __salt__["jboss7.update_datasource"](
                jboss_config=jboss_config,
                name=name,
                new_properties=datasource_properties,
                profile=profile,
            )
            if not update_result["success"]:
                ret["result"] = False
                ret["comment"] = (
                    "Could not update datasource. " + update_result["comment"]
                )
                # some changes to the datasource may have already been made, therefore we don't quit here
            else:
                ret["comment"] = "Datasource updated."

            read_result = __salt__["jboss7.read_datasource"](
                jboss_config=jboss_config, name=name, profile=profile
            )
            ds_new_properties = read_result["result"]
    else:
        if ds_result["err_code"] in (
            "JBAS014807",
            "WFLYCTL0216",
        ):  # ok, resource not exists:
            create_result = __salt__["jboss7.create_datasource"](
                jboss_config=jboss_config,
                name=name,
                datasource_properties=datasource_properties,
                profile=profile,
            )
            if create_result["success"]:
                read_result = __salt__["jboss7.read_datasource"](
                    jboss_config=jboss_config, name=name, profile=profile
                )
                ds_new_properties = read_result["result"]
                ret["comment"] = "Datasource created."
            else:
                ret["result"] = False
                ret["comment"] = (
                    "Could not create datasource. Stdout: " + create_result["stdout"]
                )
        else:
            raise CommandExecutionError(
                "Unable to handle error: {}".format(ds_result["failure-description"])
            )

    if ret["result"]:
        log.debug("ds_new_properties=%s", ds_new_properties)
        log.debug("ds_current_properties=%s", ds_current_properties)
        diff = dictdiffer.diff(ds_new_properties, ds_current_properties)

        added = diff.added()
        if len(added) > 0:
            has_changed = True
            ret["changes"]["added"] = __format_ds_changes(
                added, ds_current_properties, ds_new_properties
            )

        removed = diff.removed()
        if len(removed) > 0:
            has_changed = True
            ret["changes"]["removed"] = __format_ds_changes(
                removed, ds_current_properties, ds_new_properties
            )

        changed = diff.changed()
        if len(changed) > 0:
            has_changed = True
            ret["changes"]["changed"] = __format_ds_changes(
                changed, ds_current_properties, ds_new_properties
            )

        if not has_changed:
            ret["comment"] = "Datasource not changed."

    return ret


def __format_ds_changes(keys, old_dict, new_dict):
    log.debug(
        "__format_ds_changes(keys=%s, old_dict=%s, new_dict=%s)",
        keys,
        old_dict,
        new_dict,
    )
    changes = ""
    for key in keys:
        log.debug("key=%s", key)
        if key in old_dict and key in new_dict:
            changes += (
                key
                + ":"
                + __get_ds_value(old_dict, key)
                + "->"
                + __get_ds_value(new_dict, key)
                + "\n"
            )
        elif key in old_dict:
            changes += key + "\n"
        elif key in new_dict:
            changes += key + ":" + __get_ds_value(new_dict, key) + "\n"
    return changes


def __get_ds_value(dct, key):
    log.debug("__get_value(dict,%s)", key)
    if key == "password":
        return "***"
    elif dct[key] is None:
        return "undefined"
    else:
        return str(dct[key])


def bindings_exist(name, jboss_config, bindings, profile=None):
    """
    Ensures that given JNDI binding are present on the server.
    If a binding doesn't exist on the server it will be created.
    If it already exists its value will be changed.

    jboss_config:
        Dict with connection properties (see state description)
    bindings:
        Dict with bindings to set.
    profile:
        The profile name (domain mode only)

    Example:

    .. code-block:: yaml

            jndi_entries_created:
              jboss7.bindings_exist:
               - bindings:
                  'java:global/sampleapp/environment': 'DEV'
                  'java:global/sampleapp/configurationFile': '/var/opt/sampleapp/config.properties'
               - jboss_config: {{ pillar['jboss'] }}

    """
    log.debug(
        " ======================== STATE: jboss7.bindings_exist (name: %s) (profile:"
        " %s) ",
        name,
        profile,
    )
    log.debug("bindings=%s", bindings)
    ret = {
        "name": name,
        "result": True,
        "changes": {},
        "comment": "Bindings not changed.",
    }

    has_changed = False
    for key in bindings:
        value = str(bindings[key])
        query_result = __salt__["jboss7.read_simple_binding"](
            binding_name=key, jboss_config=jboss_config, profile=profile
        )
        if query_result["success"]:
            current_value = query_result["result"]["value"]
            if current_value != value:
                update_result = __salt__["jboss7.update_simple_binding"](
                    binding_name=key,
                    value=value,
                    jboss_config=jboss_config,
                    profile=profile,
                )
                if update_result["success"]:
                    has_changed = True
                    __log_binding_change(
                        ret["changes"], "changed", key, value, current_value
                    )
                else:
                    raise CommandExecutionError(update_result["failure-description"])
        else:
            if query_result["err_code"] in (
                "JBAS014807",
                "WFLYCTL0216",
            ):  # ok, resource not exists:
                create_result = __salt__["jboss7.create_simple_binding"](
                    binding_name=key,
                    value=value,
                    jboss_config=jboss_config,
                    profile=profile,
                )
                if create_result["success"]:
                    has_changed = True
                    __log_binding_change(ret["changes"], "added", key, value)
                else:
                    raise CommandExecutionError(create_result["failure-description"])
            else:
                raise CommandExecutionError(query_result["failure-description"])

    if has_changed:
        ret["comment"] = "Bindings changed."
    return ret


def __log_binding_change(changes, type_, key, new, old=None):
    if type_ not in changes:
        changes[type_] = ""
    if old is None:
        changes[type_] += key + ":" + new + "\n"
    else:
        changes[type_] += key + ":" + old + "->" + new + "\n"


def deployed(name, jboss_config, salt_source=None):
    """Ensures that the given application is deployed on server.

    jboss_config:
        Dict with connection properties (see state description)
    salt_source:
        How to find the artifact to be deployed.
            target_file:
                Where to look in the minion's file system for the artifact to be deployed (e.g. '/tmp/application-web-0.39.war').  When source is specified,  also specifies where to save the retrieved file.
            source:
                (optional) File on salt master (e.g. salt://application-web-0.39.war).  If absent, no files will be retrieved and the artifact in target_file will be used for the deployment.
            undeploy:
                (optional) Regular expression to match against existing deployments.  When present, if there is a deployment that matches the regular expression, it will be undeployed before the new artifact is deployed.
            undeploy_force:
                (optional) If True, the artifact will be undeployed although it has not changed.

    Examples:

    Deployment of a file from minion's local file system:

    .. code-block:: yaml

        application_deployed:
          jboss7.deployed:
            - salt_source:
                target_file: '/tmp/webapp.war'
            - jboss_config: {{ pillar['jboss'] }}

    It is assumed that /tmp/webapp.war was made available by some
    other means.  No applications will be undeployed; if an existing
    deployment that shares that name exists, then it will be replaced
    with the updated version.

    Deployment of a file from the Salt master's file system:

    .. code-block:: yaml

        application_deployed:
          jboss7.deployed:
           - salt_source:
                source: salt://application-web-0.39.war
                target_file: '/tmp/application-web-0.39.war'
                undeploy: 'application-web-.*'
           - jboss_config: {{ pillar['jboss'] }}

    Here, application-web-0.39.war file is downloaded from Salt file system to /tmp/application-web-0.39.war file on minion.
    Existing deployments are checked if any of them matches 'application-web-.*' regular expression, and if so then it
    is undeployed before deploying the application. This is useful to automate deployment of new application versions.

    If the source parameter of salt_source is specified, it can use
    any protocol that the file states use.  This includes not only
    downloading from the master but also HTTP, HTTPS, FTP,
    Amazon S3, and OpenStack Swift.

    """
    log.debug(" ======================== STATE: jboss7.deployed (name: %s) ", name)
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    comment = ""

    validate_success, validate_comment = __validate_arguments(jboss_config, salt_source)
    if not validate_success:
        return _error(ret, validate_comment)

    resolved_source, get_artifact_comment, changed = __get_artifact(salt_source)
    log.debug("resolved_source=%s", resolved_source)
    log.debug("get_artifact_comment=%s", get_artifact_comment)

    comment = __append_comment(
        new_comment=get_artifact_comment, current_comment=comment
    )
    if resolved_source is None:
        return _error(ret, get_artifact_comment)

    find_success, deployment, find_comment = __find_deployment(
        jboss_config, salt_source
    )
    if not find_success:
        return _error(ret, find_comment)

    require_deployment = True

    log.debug("deployment=%s", deployment)
    if deployment is not None:
        if "undeploy_force" in salt_source:
            if salt_source["undeploy_force"]:
                ret["changes"]["undeployed"] = __undeploy(jboss_config, deployment)
            else:
                if changed:
                    ret["changes"]["undeployed"] = __undeploy(jboss_config, deployment)
                else:
                    require_deployment = False
                    comment = __append_comment(
                        new_comment="The artifact {} was already deployed".format(
                            deployment
                        ),
                        current_comment=comment,
                    )
        else:
            ret["changes"]["undeployed"] = __undeploy(jboss_config, deployment)

    if require_deployment:
        deploy_result = __salt__["jboss7.deploy"](
            jboss_config=jboss_config, source_file=resolved_source
        )
        log.debug("deploy_result=%s", str(deploy_result))
        if deploy_result["success"]:
            comment = __append_comment(
                new_comment="Deployment completed.", current_comment=comment
            )
            ret["changes"]["deployed"] = resolved_source
        else:
            comment = __append_comment(
                new_comment="""Deployment failed\nreturn code={retcode}\nstdout='{stdout}'\nstderr='{stderr}""".format(
                    **deploy_result
                ),
                current_comment=comment,
            )
            _error(ret, comment)

    ret["comment"] = comment

    return ret


def __undeploy(jboss_config, deployment):
    __salt__["jboss7.undeploy"](jboss_config, deployment)
    return deployment


def __validate_arguments(jboss_config, salt_source):
    result, comment = __check_dict_contains(
        jboss_config, "jboss_config", ["cli_path", "controller"]
    )
    if salt_source is None:
        result = False
        comment = __append_comment("No salt_source defined", comment)
    result, comment = __check_dict_contains(
        salt_source, "salt_source", ["target_file"], comment, result
    )
    return result, comment


def __find_deployment(jboss_config, salt_source=None):
    result = None
    success = True
    comment = ""
    deployments = __salt__["jboss7.list_deployments"](jboss_config)
    if (
        salt_source is not None
        and "undeploy" in salt_source
        and salt_source["undeploy"]
    ):
        deployment_re = re.compile(salt_source["undeploy"])
        for deployment in deployments:
            if deployment_re.match(deployment):
                if result is not None:
                    success = False
                    comment = (
                        "More than one deployment matches regular expression: {}. \nFor"
                        " deployments from Salt file system deployments on JBoss are"
                        " searched to find one that matches regular expression in"
                        " 'undeploy' parameter.\nExisting deployments: {}".format(
                            salt_source["undeploy"], ",".join(deployments)
                        )
                    )
                else:
                    result = deployment

    return success, result, comment


def __get_artifact(salt_source):
    resolved_source = None
    comment = None
    changed = False

    if salt_source is None:
        log.debug("salt_source == None")
        comment = "No salt_source defined"

    elif isinstance(salt_source, dict):
        log.debug("file from salt master")

        if "source" in salt_source:
            try:
                sfn, source_sum, comment_ = __salt__["file.get_managed"](
                    name=salt_source["target_file"],
                    template=None,
                    source=salt_source["source"],
                    source_hash=None,
                    source_hash_name=None,
                    user=None,
                    group=None,
                    mode=None,
                    attrs=None,
                    saltenv=__env__,
                    context=None,
                    defaults=None,
                    skip_verify=False,
                    kwargs=None,
                )

                manage_result = __salt__["file.manage_file"](
                    name=salt_source["target_file"],
                    sfn=sfn,
                    ret=None,
                    source=salt_source["source"],
                    source_sum=source_sum,
                    user=None,
                    group=None,
                    mode=None,
                    attrs=None,
                    saltenv=__env__,
                    backup=None,
                    makedirs=False,
                    template=None,
                    show_diff=True,
                    contents=None,
                    dir_mode=None,
                )

                if manage_result["result"]:
                    resolved_source = salt_source["target_file"]
                else:
                    comment = manage_result["comment"]

                if manage_result["changes"]:
                    changed = True

            except Exception as e:  # pylint: disable=broad-except
                log.debug(traceback.format_exc())
                comment = f"Unable to manage file: {e}"

        else:
            resolved_source = salt_source["target_file"]
            comment = ""

    return resolved_source, comment, changed


def reloaded(name, jboss_config, timeout=60, interval=5):
    """
    Reloads configuration of jboss server.

    jboss_config:
        Dict with connection properties (see state description)
    timeout:
        Time to wait until jboss is back in running state. Default timeout is 60s.
    interval:
        Interval between state checks. Default interval is 5s. Decreasing the interval may slightly decrease waiting time
        but be aware that every status check is a call to jboss-cli which is a java process. If interval is smaller than
        process cleanup time it may easily lead to excessive resource consumption.

    This step performs the following operations:

    * Ensures that server is in running or reload-required state (by reading server-state attribute)
    * Reloads configuration
    * Waits for server to reload and be in running state

    Example:

    .. code-block:: yaml

        configuration_reloaded:
           jboss7.reloaded:
            - jboss_config: {{ pillar['jboss'] }}
    """
    log.debug(" ======================== STATE: jboss7.reloaded (name: %s) ", name)
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    status = __salt__["jboss7.status"](jboss_config)
    if not status["success"] or status["result"] not in ("running", "reload-required"):
        ret["result"] = False
        ret["comment"] = (
            "Cannot reload server configuration, it should be up and in 'running' or"
            " 'reload-required' state."
        )
        return ret

    result = __salt__["jboss7.reload"](jboss_config)
    if (
        result["success"]
        or "Operation failed: Channel closed" in result["stdout"]
        or "Communication error: java.util.concurrent.ExecutionException: Operation failed"
        in result["stdout"]
    ):
        wait_time = 0
        status = None
        while (
            status is None or not status["success"] or status["result"] != "running"
        ) and wait_time < timeout:
            time.sleep(interval)
            wait_time += interval
            status = __salt__["jboss7.status"](jboss_config)

        if status["success"] and status["result"] == "running":
            ret["result"] = True
            ret["comment"] = "Configuration reloaded"
            ret["changes"]["reloaded"] = "configuration"
        else:
            ret["result"] = False
            ret["comment"] = (
                "Could not reload the configuration. Timeout ({} s) exceeded. ".format(
                    timeout
                )
            )
            if not status["success"]:
                ret["comment"] = __append_comment(
                    "Could not connect to JBoss controller.", ret["comment"]
                )
            else:
                ret["comment"] = __append_comment(
                    "Server is in {} state".format(status["result"]), ret["comment"]
                )
    else:
        ret["result"] = False
        ret["comment"] = (
            "Could not reload the configuration, stdout:" + result["stdout"]
        )

    return ret


def __check_dict_contains(dct, dict_name, keys, comment="", result=True):
    for key in keys:
        if key not in dct.keys():
            result = False
            comment = __append_comment(f"Missing {key} in {dict_name}", comment)
    return result, comment


def __append_comment(new_comment, current_comment=""):
    if current_comment is None and new_comment is None:
        return ""
    if current_comment is None:
        return new_comment
    if new_comment is None:
        return current_comment
    return current_comment + "\n" + new_comment


def _error(ret, err_msg):
    ret["result"] = False
    ret["comment"] = err_msg
    return ret



"""
Management of zc.buildout

.. versionadded:: 2014.1.0

.. _`minitage's buildout maker`: https://github.com/minitage/minitage/blob/master/src/minitage/core/makers/buildout.py

This module is inspired by `minitage's buildout maker`_

.. note::

    The zc.buildout integration is still in beta; the API is subject to change

General notes
-------------

You have those following methods:

* upgrade_bootstrap
* bootstrap
* run_buildout
* buildout
"""

import copy
import logging
import os
import re
import sys
import traceback
import urllib.request

import salt.utils.files
import salt.utils.path
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError

INVALID_RESPONSE = "Unexpected response from buildout"
VALID_RESPONSE = ""
NOTSET = object()
HR = "{}\n".format("-" * 80)
RE_F = re.S | re.M | re.U
BASE_STATUS = {
    "status": None,
    "logs": [],
    "comment": "",
    "out": None,
    "logs_by_level": {},
    "outlog": None,
    "outlog_by_level": None,
}
_URL_VERSIONS = {
    1: "http://downloads.buildout.org/1/bootstrap.py",
    2: "http://downloads.buildout.org/2/bootstrap.py",
}
DEFAULT_VER = 2
_logger = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "buildout"


def __virtual__():
    """
    Only load if buildout libs are present
    """
    return __virtualname__


def _salt_callback(func, **kwargs):
    LOG.clear()

    def _call_callback(*a, **kw):
        # cleanup the module kwargs before calling it from the
        # decorator
        kw = copy.deepcopy(kw)
        for k in [ar for ar in kw if "__pub" in ar]:
            kw.pop(k, None)
        st = BASE_STATUS.copy()
        directory = kw.get("directory", ".")
        onlyif = kw.get("onlyif", None)
        unless = kw.get("unless", None)
        runas = kw.get("runas", None)
        env = kw.get("env", ())
        status = BASE_STATUS.copy()
        try:
            # may rise _ResultTransmission
            status = _check_onlyif_unless(
                onlyif, unless, directory=directory, runas=runas, env=env
            )
            # if onlyif/unless returns, we are done
            if status is None:
                status = BASE_STATUS.copy()
                comment, st = "", True
                out = func(*a, **kw)
                # we may have already final statuses not to be touched
                # merged_statuses flag is there to check that !
                if not isinstance(out, dict):
                    status = _valid(status, out=out)
                else:
                    if out.get("merged_statuses", False):
                        status = out
                    else:
                        status = _set_status(
                            status,
                            status=out.get("status", True),
                            comment=out.get("comment", ""),
                            out=out.get("out", out),
                        )
        except Exception:  # pylint: disable=broad-except
            trace = traceback.format_exc()
            LOG.error(trace)
            _invalid(status)
        LOG.clear()
        # before returning, trying to compact the log output
        for k in ["comment", "out", "outlog"]:
            if status[k] and isinstance(status[k], str):
                status[k] = "\n".join(
                    [log for log in status[k].split("\n") if log.strip()]
                )
        return status

    _call_callback.__doc__ = func.__doc__
    return _call_callback


class _Logger:
    levels = ("info", "warn", "debug", "error")

    def __init__(self):
        self._msgs = []
        self._by_level = {}

    def _log(self, level, msg):
        if not isinstance(msg, str):
            msg = msg.decode("utf-8")
        if level not in self._by_level:
            self._by_level[level] = []
        self._msgs.append((level, msg))
        self._by_level[level].append(msg)

    def debug(self, msg):
        self._log("debug", msg)

    def info(self, msg):
        self._log("info", msg)

    def error(self, msg):
        self._log("error", msg)

    def warn(self, msg):
        self._log("warn", msg)

    warning = warn

    def clear(self):
        for i in self._by_level:
            self._by_level[i] = []
        for i in self._msgs[:]:
            self._msgs.pop()

    def get_logs(self, level):
        return self._by_level.get(level, [])

    @property
    def messages(self):
        return self._msgs

    @property
    def by_level(self):
        return self._by_level


LOG = _Logger()


def _encode_status(status):
    if status["out"] is None:
        status["out"] = None
    else:
        status["out"] = salt.utils.stringutils.to_unicode(status["out"])
    status["outlog_by_level"] = salt.utils.stringutils.to_unicode(
        status["outlog_by_level"]
    )
    if status["logs"]:
        for i, data in enumerate(status["logs"][:]):
            status["logs"][i] = (data[0], salt.utils.stringutils.to_unicode(data[1]))
        for logger in "error", "warn", "info", "debug":
            logs = status["logs_by_level"].get(logger, [])[:]
            if logs:
                for i, log in enumerate(logs):
                    status["logs_by_level"][logger][i] = (
                        salt.utils.stringutils.to_unicode(log)
                    )
    return status


def _set_status(m, comment=INVALID_RESPONSE, status=False, out=None):
    """
    Assign status data to a dict.
    """
    m["out"] = out
    m["status"] = status
    m["logs"] = LOG.messages[:]
    m["logs_by_level"] = LOG.by_level.copy()
    outlog, outlog_by_level = "", ""
    m["comment"] = comment
    if out and isinstance(out, str):
        outlog += HR
        outlog += "OUTPUT:\n"
        outlog += f"{salt.utils.stringutils.to_unicode(out)}\n"
        outlog += HR
    if m["logs"]:
        outlog += HR
        outlog += "Log summary:\n"
        outlog += HR
        outlog_by_level += HR
        outlog_by_level += "Log summary by level:\n"
        outlog_by_level += HR
        for level, msg in m["logs"]:
            outlog += "\n{}: {}\n".format(
                level.upper(), salt.utils.stringutils.to_unicode(msg)
            )
        for logger in "error", "warn", "info", "debug":
            logs = m["logs_by_level"].get(logger, [])
            if logs:
                outlog_by_level += f"\n{logger.upper()}:\n"
                for idx, log in enumerate(logs[:]):
                    logs[idx] = salt.utils.stringutils.to_unicode(log)
                outlog_by_level += "\n".join(logs)
                outlog_by_level += "\n"
        outlog += HR
    m["outlog"] = outlog
    m["outlog_by_level"] = outlog_by_level
    return _encode_status(m)


def _invalid(m, comment=INVALID_RESPONSE, out=None):
    """
    Return invalid status.
    """
    return _set_status(m, status=False, comment=comment, out=out)


def _valid(m, comment=VALID_RESPONSE, out=None):
    """
    Return valid status.
    """
    return _set_status(m, status=True, comment=comment, out=out)


def _Popen(
    command,
    output=False,
    directory=".",
    runas=None,
    env=(),
    exitcode=0,
    use_vt=False,
    loglevel=None,
):
    """
    Run a command.

    output
        return output if true

    directory
        directory to execute in

    runas
        user used to run buildout as

    env
        environment variables to set when running

    exitcode
        fails if cmd does not return this exit code
        (set to None to disable check)

    use_vt
        Use the new salt VT to stream output [experimental]

    """
    ret = None
    directory = os.path.abspath(directory)
    if isinstance(command, list):
        command = " ".join(command)
    LOG.debug(f"Running {command}")
    if not loglevel:
        loglevel = "debug"
    ret = __salt__["cmd.run_all"](
        command,
        cwd=directory,
        output_loglevel=loglevel,
        runas=runas,
        env=env,
        use_vt=use_vt,
        python_shell=False,
    )
    out = ret["stdout"] + "\n\n" + ret["stderr"]
    if (exitcode is not None) and (ret["retcode"] != exitcode):
        raise _BuildoutError(out)
    ret["output"] = out
    if output:
        ret = out
    return ret


class _BuildoutError(CommandExecutionError):
    """
    General Buildout Error.
    """


def _has_old_distribute(python=sys.executable, runas=None, env=()):
    old_distribute = False
    try:
        cmd = [
            python,
            "-c",
            "'import pkg_resources;"
            "print pkg_resources."
            'get_distribution("distribute").location\'',
        ]
        ret = _Popen(cmd, runas=runas, env=env, output=True)
        if "distribute-0.6" in ret:
            old_distribute = True
    except Exception:  # pylint: disable=broad-except
        old_distribute = False
    return old_distribute


def _has_setuptools7(python=sys.executable, runas=None, env=()):
    new_st = False
    try:
        cmd = [
            python,
            "-c",
            "'import pkg_resources;"
            "print not pkg_resources."
            'get_distribution("setuptools").version.startswith("0.6")\'',
        ]
        ret = _Popen(cmd, runas=runas, env=env, output=True)
        if "true" in ret.lower():
            new_st = True
    except Exception:  # pylint: disable=broad-except
        new_st = False
    return new_st


def _find_cfgs(path, cfgs=None):
    """
    Find all buildout configs in a subdirectory.
    only buildout.cfg and etc/buildout.cfg are valid in::

    path
        directory where to start to search

    cfg
        a optional list to append to

            .
            ├── buildout.cfg
            ├── etc
            │   └── buildout.cfg
            ├── foo
            │   └── buildout.cfg
            └── var
                └── buildout.cfg
    """
    ignored = ["var", "parts"]
    dirs = []
    if not cfgs:
        cfgs = []
    for i in os.listdir(path):
        fi = os.path.join(path, i)
        if fi.endswith(".cfg") and os.path.isfile(fi):
            cfgs.append(fi)
        if os.path.isdir(fi) and (i not in ignored):
            dirs.append(fi)
    for fpath in dirs:
        for p, ids, ifs in salt.utils.path.os_walk(fpath):
            for i in ifs:
                if i.endswith(".cfg"):
                    cfgs.append(os.path.join(p, i))
    return cfgs


def _get_bootstrap_content(directory="."):
    """
    Get the current bootstrap.py script content
    """
    try:
        with salt.utils.files.fopen(
            os.path.join(os.path.abspath(directory), "bootstrap.py")
        ) as fic:
            oldcontent = salt.utils.stringutils.to_unicode(fic.read())
    except OSError:
        oldcontent = ""
    return oldcontent


def _get_buildout_ver(directory="."):
    """Check for buildout versions.

    In any cases, check for a version pinning
    Also check for buildout.dumppickedversions which is buildout1 specific
    Also check for the version targeted by the local bootstrap file
    Take as default buildout2

    directory
        directory to execute in
    """
    directory = os.path.abspath(directory)
    buildoutver = 2
    try:
        files = _find_cfgs(directory)
        for f in files:
            with salt.utils.files.fopen(f) as fic:
                buildout1re = re.compile(r"^zc\.buildout\s*=\s*1", RE_F)
                dfic = salt.utils.stringutils.to_unicode(fic.read())
                if ("buildout.dumppick" in dfic) or (buildout1re.search(dfic)):
                    buildoutver = 1
        bcontent = _get_bootstrap_content(directory)
        if (
            "--download-base" in bcontent
            or "--setup-source" in bcontent
            or "--distribute" in bcontent
        ):
            buildoutver = 1
    except OSError:
        pass
    return buildoutver


def _get_bootstrap_url(directory):
    """
    Get the most appropriate download URL for the bootstrap script.

    directory
        directory to execute in

    """
    v = _get_buildout_ver(directory)
    return _URL_VERSIONS.get(v, _URL_VERSIONS[DEFAULT_VER])


def _dot_buildout(directory):
    """
    Get the local marker directory.

    directory
        directory to execute in
    """
    return os.path.join(os.path.abspath(directory), ".buildout")


@_salt_callback
def upgrade_bootstrap(
    directory=".",
    onlyif=None,
    unless=None,
    runas=None,
    env=(),
    offline=False,
    buildout_ver=None,
):
    """
    Upgrade current bootstrap.py with the last released one.

    Indeed, when we first run a buildout, a common source of problem
    is to have a locally stale bootstrap, we just try to grab a new copy

    directory
        directory to execute in

    offline
        are we executing buildout in offline mode

    buildout_ver
        forcing to use a specific buildout version (1 | 2)

    onlyif
        Only execute cmd if statement on the host return 0

    unless
        Do not execute cmd if statement on the host return 0

    CLI Example:

    .. code-block:: bash

        salt '*' buildout.upgrade_bootstrap /srv/mybuildout
    """
    if buildout_ver:
        booturl = _URL_VERSIONS[buildout_ver]
    else:
        buildout_ver = _get_buildout_ver(directory)
        booturl = _get_bootstrap_url(directory)
    LOG.debug(f"Using {booturl}")
    # try to download an up-to-date bootstrap
    # set defaulttimeout
    # and add possible content
    directory = os.path.abspath(directory)
    b_py = os.path.join(directory, "bootstrap.py")
    comment = ""
    try:
        oldcontent = _get_bootstrap_content(directory)
        dbuild = _dot_buildout(directory)
        data = oldcontent
        updated = False
        dled = False
        if not offline:
            try:
                if not os.path.isdir(dbuild):
                    os.makedirs(dbuild)
                # only try to download once per buildout checkout
                with salt.utils.files.fopen(
                    os.path.join(dbuild, f"{buildout_ver}.updated_bootstrap")
                ):
                    pass
            except OSError:
                LOG.info("Bootstrap updated from repository")
                data = urllib.request.urlopen(booturl).read()
                updated = True
                dled = True
        if "socket.setdefaulttimeout" not in data:
            updated = True
            ldata = data.splitlines()
            ldata.insert(1, "import socket;socket.setdefaulttimeout(2)")
            data = "\n".join(ldata)
        if updated:
            comment = "Bootstrap updated"
            with salt.utils.files.fopen(b_py, "w") as fic:
                fic.write(salt.utils.stringutils.to_str(data))
        if dled:
            with salt.utils.files.fopen(
                os.path.join(dbuild, f"{buildout_ver}.updated_bootstrap"), "w"
            ) as afic:
                afic.write("foo")
    except OSError:
        if oldcontent:
            with salt.utils.files.fopen(b_py, "w") as fic:
                fic.write(salt.utils.stringutils.to_str(oldcontent))

    return {"comment": comment}


@_salt_callback
def bootstrap(
    directory=".",
    config="buildout.cfg",
    python=sys.executable,
    onlyif=None,
    unless=None,
    runas=None,
    env=(),
    distribute=None,
    buildout_ver=None,
    test_release=False,
    offline=False,
    new_st=None,
    use_vt=False,
    loglevel=None,
):
    """
    Run the buildout bootstrap dance (python bootstrap.py).

    directory
        directory to execute in

    config
        alternative buildout configuration file to use

    runas
        User used to run buildout as

    env
        environment variables to set when running

    buildout_ver
        force a specific buildout version (1 | 2)

    test_release
        buildout accept test release

    offline
        are we executing buildout in offline mode

    distribute
        Forcing use of distribute

    new_st
        Forcing use of setuptools >= 0.7

    python
        path to a python executable to use in place of default (salt one)

    onlyif
        Only execute cmd if statement on the host return 0

    unless
        Do not execute cmd if statement on the host return 0

    use_vt
        Use the new salt VT to stream output [experimental]

    CLI Example:

    .. code-block:: bash

        salt '*' buildout.bootstrap /srv/mybuildout
    """
    directory = os.path.abspath(directory)
    dbuild = _dot_buildout(directory)
    bootstrap_args = ""
    has_distribute = _has_old_distribute(python=python, runas=runas, env=env)
    has_new_st = _has_setuptools7(python=python, runas=runas, env=env)
    if has_distribute and has_new_st and not distribute and new_st:
        new_st = True
        distribute = False
    if has_distribute and has_new_st and not distribute and new_st:
        new_st = True
        distribute = False
    if has_distribute and has_new_st and distribute and not new_st:
        new_st = True
        distribute = False
    if has_distribute and has_new_st and not distribute and not new_st:
        new_st = True
        distribute = False

    if not has_distribute and has_new_st and not distribute and new_st:
        new_st = True
        distribute = False
    if not has_distribute and has_new_st and not distribute and new_st:
        new_st = True
        distribute = False
    if not has_distribute and has_new_st and distribute and not new_st:
        new_st = True
        distribute = False
    if not has_distribute and has_new_st and not distribute and not new_st:
        new_st = True
        distribute = False

    if has_distribute and not has_new_st and not distribute and new_st:
        new_st = True
        distribute = False
    if has_distribute and not has_new_st and not distribute and new_st:
        new_st = True
        distribute = False
    if has_distribute and not has_new_st and distribute and not new_st:
        new_st = False
        distribute = True
    if has_distribute and not has_new_st and not distribute and not new_st:
        new_st = False
        distribute = True

    if not has_distribute and not has_new_st and not distribute and new_st:
        new_st = True
        distribute = False
    if not has_distribute and not has_new_st and not distribute and new_st:
        new_st = True
        distribute = False
    if not has_distribute and not has_new_st and distribute and not new_st:
        new_st = False
        distribute = True
    if not has_distribute and not has_new_st and not distribute and not new_st:
        new_st = True
        distribute = False

    if new_st and distribute:
        distribute = False
    if new_st:
        distribute = False
        LOG.warning("Forcing to use setuptools as we have setuptools >= 0.7")
    if distribute:
        new_st = False
        if buildout_ver == 1:
            LOG.warning("Using distribute !")
            bootstrap_args += " --distribute"
    if not os.path.isdir(dbuild):
        os.makedirs(dbuild)
    upgrade_bootstrap(directory, offline=offline, buildout_ver=buildout_ver)
    # be sure which buildout bootstrap we have
    b_py = os.path.join(directory, "bootstrap.py")
    with salt.utils.files.fopen(b_py) as fic:
        content = salt.utils.stringutils.to_unicode(fic.read())
    if (test_release is not False) and " --accept-buildout-test-releases" in content:
        bootstrap_args += " --accept-buildout-test-releases"
    if config and '"-c"' in content:
        bootstrap_args += f" -c {config}"
    # be sure that the bootstrap belongs to the running user
    try:
        if runas:
            uid = __salt__["user.info"](runas)["uid"]
            gid = __salt__["user.info"](runas)["gid"]
            os.chown("bootstrap.py", uid, gid)
    except OSError as exc:
        # don't block here, try to execute it if can pass
        _logger.error(
            "BUILDOUT bootstrap permissions error: %s",
            exc,
            exc_info=_logger.isEnabledFor(logging.DEBUG),
        )
    cmd = f"{python} bootstrap.py {bootstrap_args}"
    ret = _Popen(
        cmd, directory=directory, runas=runas, loglevel=loglevel, env=env, use_vt=use_vt
    )
    output = ret["output"]
    return {"comment": cmd, "out": output}


@_salt_callback
def run_buildout(
    directory=".",
    config="buildout.cfg",
    parts=None,
    onlyif=None,
    unless=None,
    offline=False,
    newest=True,
    runas=None,
    env=(),
    verbose=False,
    debug=False,
    use_vt=False,
    loglevel=None,
):
    """
    Run a buildout in a directory.

    directory
        directory to execute in

    config
        alternative buildout configuration file to use

    offline
        are we executing buildout in offline mode

    runas
        user used to run buildout as

    env
        environment variables to set when running

    onlyif
        Only execute cmd if statement on the host return 0

    unless
        Do not execute cmd if statement on the host return 0

    newest
        run buildout in newest mode

    force
        run buildout unconditionally

    verbose
        run buildout in verbose mode (-vvvvv)

    use_vt
        Use the new salt VT to stream output [experimental]

    CLI Example:

    .. code-block:: bash

        salt '*' buildout.run_buildout /srv/mybuildout
    """
    directory = os.path.abspath(directory)
    bcmd = os.path.join(directory, "bin", "buildout")
    installed_cfg = os.path.join(directory, ".installed.cfg")
    argv = []
    if verbose:
        LOG.debug("Buildout is running in verbose mode!")
        argv.append("-vvvvvvv")
    if not newest and os.path.exists(installed_cfg):
        LOG.debug("Buildout is running in non newest mode!")
        argv.append("-N")
    if newest:
        LOG.debug("Buildout is running in newest mode!")
        argv.append("-n")
    if offline:
        LOG.debug("Buildout is running in offline mode!")
        argv.append("-o")
    if debug:
        LOG.debug("Buildout is running in debug mode!")
        argv.append("-D")
    cmds, outputs = [], []
    if parts:
        for part in parts:
            LOG.info(f"Installing single part: {part}")
            cmd = "{} -c {} {} install {}".format(bcmd, config, " ".join(argv), part)
            cmds.append(cmd)
            outputs.append(
                _Popen(
                    cmd,
                    directory=directory,
                    runas=runas,
                    env=env,
                    output=True,
                    loglevel=loglevel,
                    use_vt=use_vt,
                )
            )
    else:
        LOG.info("Installing all buildout parts")
        cmd = "{} -c {} {}".format(bcmd, config, " ".join(argv))
        cmds.append(cmd)
        outputs.append(
            _Popen(
                cmd,
                directory=directory,
                runas=runas,
                loglevel=loglevel,
                env=env,
                output=True,
                use_vt=use_vt,
            )
        )

    return {"comment": "\n".join(cmds), "out": "\n".join(outputs)}


def _merge_statuses(statuses):
    status = BASE_STATUS.copy()
    status["status"] = None
    status["merged_statuses"] = True
    status["out"] = ""
    for st in statuses:
        if status["status"] is not False:
            status["status"] = st["status"]
        out = st["out"]
        comment = salt.utils.stringutils.to_unicode(st["comment"])
        logs = st["logs"]
        logs_by_level = st["logs_by_level"]
        outlog_by_level = st["outlog_by_level"]
        outlog = st["outlog"]
        if out:
            if not status["out"]:
                status["out"] = ""
            status["out"] += "\n"
            status["out"] += HR
            out = salt.utils.stringutils.to_unicode(out)
            status["out"] += f"{out}\n"
            status["out"] += HR
        if comment:
            if not status["comment"]:
                status["comment"] = ""
            status["comment"] += "\n{}\n".format(
                salt.utils.stringutils.to_unicode(comment)
            )
        if outlog:
            if not status["outlog"]:
                status["outlog"] = ""
            outlog = salt.utils.stringutils.to_unicode(outlog)
            status["outlog"] += f"\n{HR}"
            status["outlog"] += outlog
        if outlog_by_level:
            if not status["outlog_by_level"]:
                status["outlog_by_level"] = ""
            status["outlog_by_level"] += f"\n{HR}"
            status["outlog_by_level"] += salt.utils.stringutils.to_unicode(
                outlog_by_level
            )
        status["logs"].extend(
            [(a[0], salt.utils.stringutils.to_unicode(a[1])) for a in logs]
        )
        for log in logs_by_level:
            if log not in status["logs_by_level"]:
                status["logs_by_level"][log] = []
            status["logs_by_level"][log].extend(
                [salt.utils.stringutils.to_unicode(a) for a in logs_by_level[log]]
            )
    return _encode_status(status)


@_salt_callback
def buildout(
    directory=".",
    config="buildout.cfg",
    parts=None,
    runas=None,
    env=(),
    buildout_ver=None,
    test_release=False,
    distribute=None,
    new_st=None,
    offline=False,
    newest=False,
    python=sys.executable,
    debug=False,
    verbose=False,
    onlyif=None,
    unless=None,
    use_vt=False,
    loglevel=None,
):
    """
    Run buildout in a directory.

    directory
        directory to execute in

    config
        buildout config to use

    parts
        specific buildout parts to run

    runas
        user used to run buildout as

    env
        environment variables to set when running

    buildout_ver
        force a specific buildout version (1 | 2)

    test_release
        buildout accept test release

    new_st
        Forcing use of setuptools >= 0.7

    distribute
        use distribute over setuptools if possible

    offline
        does buildout run offline

    python
        python to use

    debug
        run buildout with -D debug flag

    onlyif
        Only execute cmd if statement on the host return 0

    unless
        Do not execute cmd if statement on the host return 0
    newest
        run buildout in newest mode

    verbose
        run buildout in verbose mode (-vvvvv)

    use_vt
        Use the new salt VT to stream output [experimental]

    CLI Example:

    .. code-block:: bash

        salt '*' buildout.buildout /srv/mybuildout
    """
    LOG.info(f"Running buildout in {directory} ({config})")
    boot_ret = bootstrap(
        directory,
        config=config,
        buildout_ver=buildout_ver,
        test_release=test_release,
        offline=offline,
        new_st=new_st,
        env=env,
        runas=runas,
        distribute=distribute,
        python=python,
        use_vt=use_vt,
        loglevel=loglevel,
    )
    buildout_ret = run_buildout(
        directory=directory,
        config=config,
        parts=parts,
        offline=offline,
        newest=newest,
        runas=runas,
        env=env,
        verbose=verbose,
        debug=debug,
        use_vt=use_vt,
        loglevel=loglevel,
    )
    # signal the decorator or our return
    return _merge_statuses([boot_ret, buildout_ret])


def _check_onlyif_unless(onlyif, unless, directory, runas=None, env=()):
    ret = None
    status = BASE_STATUS.copy()
    if os.path.exists(directory):
        directory = os.path.abspath(directory)
        status["status"] = False
        retcode = __salt__["cmd.retcode"]
        if onlyif is not None:
            if not isinstance(onlyif, str):
                if not onlyif:
                    _valid(status, "onlyif condition is false")
            elif isinstance(onlyif, str):
                if retcode(onlyif, cwd=directory, runas=runas, env=env) != 0:
                    _valid(status, "onlyif condition is false")
        if unless is not None:
            if not isinstance(unless, str):
                if unless:
                    _valid(status, "unless condition is true")
            elif isinstance(unless, str):
                if (
                    retcode(
                        unless, cwd=directory, runas=runas, env=env, python_shell=False
                    )
                    == 0
                ):
                    _valid(status, "unless condition is true")
    if status["status"]:
        ret = status
    return ret



"""
Loop state

Allows for looping over execution modules.

.. versionadded:: 2017.7.0

In both examples below, the execution module function ``boto_elb.get_instance_health``
returns a list of dicts. The condition checks the ``state``-key of the first dict
in the returned list and compares its value to the string `InService`.

.. code-block:: yaml

    wait_for_service_to_be_healthy:
      loop.until:
        - name: boto_elb.get_instance_health
        - condition: m_ret[0]['state'] == 'InService'
        - period: 5
        - timeout: 20
        - m_args:
          - {{ elb }}
        - m_kwargs:
            keyid: {{ access_key }}
            key: {{ secret_key }}
            instances: "{{ instance }}"

.. warning::

    This state allows arbitrary python code to be executed through the condition
    parameter which is literally evaluated within the state. Please use caution.

.. versionchanged:: 3000

A version that does not use eval is now available. It uses either the python ``operator``
to compare the result of the function called in ``name``, which can be one of the
following: lt, le, eq (default), ne, ge, gt.
Alternatively, `compare_operator` can be filled with a function from an execution
module in ``__salt__`` or ``__utils__`` like the example below.
The function :py:func:`data.subdict_match <salt.utils.data.subdict_match>` checks if the
``expected`` expression matches the data returned by calling the ``name`` function
(with passed ``args`` and ``kwargs``).

.. code-block:: yaml

    Wait for service to be healthy:
      loop.until_no_eval:
        - name: boto_elb.get_instance_health
        - expected: '0:state:InService'
        - compare_operator: data.subdict_match
        - period: 5
        - timeout: 20
        - args:
          - {{ elb }}
        - kwargs:
            keyid: {{ access_key }}
            key: {{ secret_key }}
            instances: "{{ instance }}"
"""

import logging
import operator
import sys
import time

# Initialize logging
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "loop"


def __virtual__():
    return True


def until(name, m_args=None, m_kwargs=None, condition=None, period=1, timeout=60):
    """
    Loop over an execution module until a condition is met.

    :param str name: The name of the execution module
    :param list m_args: The execution module's positional arguments
    :param dict m_kwargs: The execution module's keyword arguments
    :param str condition: The condition which must be met for the loop to break.
        This should contain ``m_ret`` which is the return from the execution module.
    :param period: The number of seconds to wait between executions
    :type period: int or float
    :param timeout: The timeout in seconds
    :type timeout: int or float
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if m_args is None:
        m_args = ()
    if m_kwargs is None:
        m_kwargs = {}

    if name not in __salt__:
        ret["comment"] = f"Cannot find module {name}"
    elif condition is None:
        ret["comment"] = "An exit condition must be specified"
    elif not isinstance(period, (int, float)):
        ret["comment"] = "Period must be specified as a float in seconds"
    elif not isinstance(timeout, (int, float)):
        ret["comment"] = "Timeout must be specified as a float in seconds"
    elif __opts__["test"]:
        ret["comment"] = f"The execution module {name} will be run"
        ret["result"] = None
    else:
        if m_args is None:
            m_args = []
        if m_kwargs is None:
            m_kwargs = {}

        timeout = time.time() + timeout
        while time.time() < timeout:
            m_ret = __salt__[name](*m_args, **m_kwargs)
            if eval(condition):  # pylint: disable=W0123
                ret["result"] = True
                ret["comment"] = f"Condition {condition} was met"
                break
            time.sleep(period)
        else:
            ret["comment"] = "Timed out while waiting for condition {}".format(
                condition
            )
    return ret


def until_no_eval(
    name,
    expected,
    compare_operator="eq",
    timeout=60,
    period=1,
    init_wait=0,
    args=None,
    kwargs=None,
):
    """
    Generic waiter state that waits for a specific salt function to produce an
    expected result.
    The state fails if the function does not exist or raises an exception,
    or does not produce the expected result within the allotted retries.

    :param str name: Name of the module.function to call
    :param expected: Expected return value. This can be almost anything.
    :param str compare_operator: Operator to use to compare the result of the
        module.function call with the expected value. This can be anything present
        in __salt__ or __utils__. Will be called with 2 args: result, expected.
    :param timeout: Abort after this amount of seconds (excluding init_wait).
    :type timeout: int or float
    :param period: Time (in seconds) to wait between attempts.
    :type period: int or float
    :param init_wait: Time (in seconds) to wait before trying anything.
    :type init_wait: int or float
    :param list args: args to pass to the salt module.function.
    :param dict kwargs: kwargs to pass to the salt module.function.

    .. versionadded:: 3000

    """
    ret = {"name": name, "comment": "", "changes": {}, "result": False}
    if name not in __salt__:
        ret["comment"] = f'Module.function "{name}" is unavailable.'
    elif not isinstance(period, (int, float)):
        ret["comment"] = "Period must be specified as a float in seconds"
    elif not isinstance(timeout, (int, float)):
        ret["comment"] = "Timeout must be specified as a float in seconds"
    elif compare_operator in __salt__:
        comparator = __salt__[compare_operator]
    elif compare_operator in __utils__:
        comparator = __utils__[compare_operator]
    elif not hasattr(operator, compare_operator):
        ret["comment"] = f'Invalid operator "{compare_operator}" supplied.'
    else:
        comparator = getattr(operator, compare_operator)
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = 'Would have waited for "{}" to produce "{}".'.format(
            name, expected
        )
    if ret["comment"]:
        return ret

    if init_wait:
        time.sleep(init_wait)
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    res_archive = []
    current_attempt = 0
    timeout = time.time() + timeout
    while time.time() < timeout:
        current_attempt += 1
        try:
            res = __salt__[name](*args, **kwargs)
        except Exception:  # pylint: disable=broad-except
            (exc_type, exc_value, _) = sys.exc_info()
            ret["comment"] = "Exception occurred while executing {}: {}:{}".format(
                name, exc_type, exc_value
            )
            break
        res_archive.append(res)
        cmp_res = comparator(res, expected)
        log.debug(
            "%s:until_no_eval:\n"
            "\t\tAttempt %s, result: %s, expected: %s, compare result: %s",
            __name__,
            current_attempt,
            res,
            expected,
            cmp_res,
        )
        if cmp_res:
            ret["result"] = True
            ret["comment"] = "Call provided the expected results in {} attempts".format(
                current_attempt
            )
            break
        time.sleep(period)
    else:
        ret["comment"] = (
            "Call did not produce the expected result after {} attempts".format(
                current_attempt
            )
        )
        log.debug(
            "%s:until_no_eval:\n\t\tResults of all attempts: %s",
            __name__,
            res_archive,
        )
    return ret



"""
Interaction with the Supervisor daemon
======================================

.. code-block:: yaml

    wsgi_server:
      supervisord.running:
        - require:
          - pkg: supervisor
        - watch:
          - file: /etc/nginx/sites-enabled/wsgi_server.conf
"""

import logging

log = logging.getLogger(__name__)


def _check_error(result, success_message):
    ret = {}

    if "ERROR" in result:
        if any(
            substring in result
            for substring in [
                "already started",
                "not running",
                "process group already active",
            ]
        ):
            ret["comment"] = success_message
        else:
            ret["comment"] = result
            ret["result"] = False
    else:
        ret["comment"] = success_message

    return ret


def _is_stopped_state(state):
    if state in ("STOPPED", "STOPPING", "EXITED", "FATAL", "BACKOFF"):
        return True
    if state in ("STARTING", "RUNNING"):
        return False
    return False


def running(
    name, restart=False, update=False, user=None, conf_file=None, bin_env=None, **kwargs
):
    """
    Ensure the named service is running.

    name
        Service name as defined in the supervisor configuration file

    restart
        Whether to force a restart

    update
        Whether to update the supervisor configuration.

    user
        Name of the user to run the supervisorctl command

        .. versionadded:: 0.17.0

    conf_file
        path to supervisorctl config file

    bin_env
        path to supervisorctl bin or path to virtualenv with supervisor
        installed

    """
    if name.endswith(":*"):
        name = name[:-1]

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    if "supervisord.status" not in __salt__:
        ret["result"] = False
        ret["comment"] = (
            "Supervisord module not activated. Do you need to install supervisord?"
        )
        return ret

    all_processes = __salt__["supervisord.status"](
        user=user, conf_file=conf_file, bin_env=bin_env
    )

    # parse process groups
    process_groups = set()
    for proc in all_processes:
        if ":" in proc:
            process_groups.add(proc[: proc.index(":") + 1])
    process_groups = sorted(process_groups)

    matches = {}
    if name in all_processes:
        matches[name] = all_processes[name]["state"].lower() == "running"
    elif name in process_groups:
        for process in (x for x in all_processes if x.startswith(name)):
            matches[process] = all_processes[process]["state"].lower() == "running"
    to_add = not bool(matches)

    if __opts__["test"]:
        if not to_add:
            # Process/group already present, check if any need to be started
            to_start = [x for x, y in matches.items() if y is False]
            if to_start:
                ret["result"] = None
                if name.endswith(":"):
                    # Process group
                    if len(to_start) == len(matches):
                        ret["comment"] = (
                            f"All services in group '{name}' will be started"
                        )
                    else:
                        ret["comment"] = (
                            "The following services will be started: {}".format(
                                " ".join(to_start)
                            )
                        )
                else:
                    # Single program
                    ret["comment"] = f"Service {name} will be started"
            else:
                if name.endswith(":"):
                    # Process group
                    ret["comment"] = (
                        f"All services in group '{name}' are already running"
                    )
                else:
                    ret["comment"] = f"Service {name} is already running"
        else:
            ret["result"] = None
            # Process/group needs to be added
            if name.endswith(":"):
                _type = f"Group '{name}'"
            else:
                _type = f"Service {name}"
            ret["comment"] = f"{_type} will be added and started"
        return ret

    changes = []
    just_updated = False

    if update:
        # If the state explicitly asks to update, we don't care if the process
        # is being added or not, since it'll take care of this for us,
        # so give this condition priority in order
        #
        # That is, unless `to_add` somehow manages to contain processes
        # we don't want running, in which case adding them may be a mistake
        comment = "Updating supervisor"
        result = __salt__["supervisord.update"](
            user=user, conf_file=conf_file, bin_env=bin_env
        )
        ret.update(_check_error(result, comment))
        log.debug(comment)

        if f"{name}: updated" in result:
            just_updated = True
    elif to_add:
        # Not sure if this condition is precise enough.
        comment = f"Adding service: {name}"
        __salt__["supervisord.reread"](user=user, conf_file=conf_file, bin_env=bin_env)
        # Causes supervisorctl to throw `ERROR: process group already active`
        # if process group exists. At this moment, I'm not sure how to handle
        # this outside of grepping out the expected string in `_check_error`.
        result = __salt__["supervisord.add"](
            name, user=user, conf_file=conf_file, bin_env=bin_env
        )

        ret.update(_check_error(result, comment))
        changes.append(comment)
        log.debug(comment)

    is_stopped = None

    process_type = None
    if name in process_groups:
        process_type = "group"

        # check if any processes in this group are stopped
        is_stopped = False
        for proc in all_processes:
            if proc.startswith(name) and _is_stopped_state(
                all_processes[proc]["state"]
            ):
                is_stopped = True
                break

    elif name in all_processes:
        process_type = "service"

        if _is_stopped_state(all_processes[name]["state"]):
            is_stopped = True
        else:
            is_stopped = False

    if is_stopped is False:
        if restart and not just_updated:
            comment = "Restarting{}: {}".format(
                process_type is not None and f" {process_type}" or "", name
            )
            log.debug(comment)
            result = __salt__["supervisord.restart"](
                name, user=user, conf_file=conf_file, bin_env=bin_env
            )
            ret.update(_check_error(result, comment))
            changes.append(comment)
        elif just_updated:
            comment = "Not starting updated{}: {}".format(
                process_type is not None and f" {process_type}" or "", name
            )
            result = comment
            ret.update({"comment": comment})
        else:
            comment = "Not starting already running{}: {}".format(
                process_type is not None and f" {process_type}" or "", name
            )
            result = comment
            ret.update({"comment": comment})

    elif not just_updated:
        comment = "Starting{}: {}".format(
            process_type is not None and f" {process_type}" or "", name
        )
        changes.append(comment)
        log.debug(comment)
        result = __salt__["supervisord.start"](
            name, user=user, conf_file=conf_file, bin_env=bin_env
        )

        ret.update(_check_error(result, comment))
        log.debug(str(result))

    if ret["result"] and changes:
        ret["changes"][name] = " ".join(changes)
    return ret


def dead(name, user=None, conf_file=None, bin_env=None, **kwargs):
    """
    Ensure the named service is dead (not running).

    name
        Service name as defined in the supervisor configuration file

    user
        Name of the user to run the supervisorctl command

        .. versionadded:: 0.17.0

    conf_file
        path to supervisorctl config file

    bin_env
        path to supervisorctl bin or path to virtualenv with supervisor
        installed

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Service {name} is set to be stopped"
    else:
        comment = f"Stopping service: {name}"
        log.debug(comment)

        all_processes = __salt__["supervisord.status"](
            user=user, conf_file=conf_file, bin_env=bin_env
        )

        # parse process groups
        process_groups = []
        for proc in all_processes:
            if ":" in proc:
                process_groups.append(proc[: proc.index(":") + 1])
        process_groups = list(set(process_groups))

        is_stopped = None

        if name in process_groups:
            # check if any processes in this group are stopped
            is_stopped = False
            for proc in all_processes:
                if proc.startswith(name) and _is_stopped_state(
                    all_processes[proc]["state"]
                ):
                    is_stopped = True
                    break

        elif name in all_processes:
            if _is_stopped_state(all_processes[name]["state"]):
                is_stopped = True
            else:
                is_stopped = False
        else:
            # process name doesn't exist
            ret["comment"] = f"Service {name} doesn't exist"
            return ret

        if is_stopped is True:
            ret["comment"] = f"Service {name} is not running"
        else:
            result = {
                name: __salt__["supervisord.stop"](
                    name, user=user, conf_file=conf_file, bin_env=bin_env
                )
            }
            ret.update(_check_error(result, comment))
            ret["changes"][name] = comment
            log.debug(str(result))
    return ret


def mod_watch(
    name, restart=True, update=False, user=None, conf_file=None, bin_env=None, **kwargs
):
    """
    The supervisord watcher, called to invoke the watch command.
    Always restart on watch

    .. note::
        This state exists to support special handling of the ``watch``
        :ref:`requisite <requisites>`. It should not be called directly.

        Parameters for this function should be set by the state being triggered.
    """
    return running(
        name,
        restart=restart,
        update=update,
        user=user,
        conf_file=conf_file,
        bin_env=bin_env,
    )



"""
This module contains routines used for the salt mine
"""

import logging

import salt.utils.data

log = logging.getLogger(__name__)

MINE_ITEM_ACL_ID = "__saltmine_acl__"
MINE_ITEM_ACL_VERSION = 1
MINE_ITEM_ACL_DATA = "__data__"


def minion_side_acl_denied(minion_acl_cache, mine_minion, mine_function, req_minion):
    """
    Helper function to determine if a ``req_minion`` is not allowed to retrieve
    ``mine_function``-data from the mine of ``mine_minion``.

    :param dict minion_acl_cache: Contains minion_id as first level key, and mine
        function as 2nd level key. Value of 2nd level is a list of minions that
        are allowed to retrieve the function from the mine of the minion.
    :param str mine_minion: The minion that the mine value originated from.
    :param str mine_function: The mine function that is requested.
    :param str req_minion: The minion that is requesting the mine data.

    :rtype: bool
    :return:
        False if no ACL has been defined for ``mine_minion``, ``mine_function``.
        False if an ACL has been defined and it grants access.
        True if an ACL has been defined and does not grant access.
    """
    minion_acl_entry = minion_acl_cache.get(mine_minion, {}).get(mine_function, [])
    ret = minion_acl_entry and req_minion not in minion_acl_entry
    if ret:
        log.debug(
            "Salt mine request from %s for function %s on minion %s denied.",
            req_minion,
            mine_function,
            mine_minion,
        )
    return ret


def wrap_acl_structure(function_data, allow_tgt=None, allow_tgt_type=None):
    """
    Helper function to convert an non-ACL mine entry into the new entry which
    includes ACL data.

    :param dict function_data: The function data to wrap.
    :param str allow_tgt: The targeting string that designates which minions can
        request this mine entry.
    :param str allow_tgt_type: The type of targeting string.
        .. seealso:: :ref:`targeting`

    :rtype: dict
    :return: Mine entry structured to include ACL data.
    """
    res = {
        MINE_ITEM_ACL_DATA: function_data,
        MINE_ITEM_ACL_ID: MINE_ITEM_ACL_VERSION,
    }
    # Add minion-side ACL
    if allow_tgt:
        res.update(
            salt.utils.data.filter_falsey(
                {"allow_tgt": allow_tgt, "allow_tgt_type": allow_tgt_type}
            )
        )
    return res


def parse_function_definition(function_definition):
    """
    Helper function to parse the mine_function definition as provided in config,
    or pillar.

    :param function_definition: The function definition to parse.
    :type function_definition: list or dict

    :rtype: tuple
    :return: Tuple with function_name, function_args, function_kwargs, minion_acl (dict)
    """
    function_name = None
    function_args = []
    function_kwargs = {}
    minion_acl = {}
    if isinstance(function_definition, dict):
        # dictionary format for specifying mine function
        function_name = function_definition.pop("mine_function", None)
        function_kwargs = function_definition
    elif isinstance(function_definition, list):
        for item in function_definition:
            if isinstance(item, dict):
                # if len(item) > 1: # Multiple kwargs in a single list item
                function_kwargs.update(item)
            else:
                function_args.append(item)
        function_name = function_kwargs.pop("mine_function", None)

    minion_acl = salt.utils.data.filter_falsey(
        {
            "allow_tgt": function_kwargs.pop("allow_tgt", None),
            "allow_tgt_type": function_kwargs.pop("allow_tgt_type", None),
        }
    )

    return (function_name, function_args, function_kwargs, minion_acl)
